#ifndef BMATRIX_H
#define BMATRIX_H

#include <stdio.h>
#include <math.h>

template<class T>
class BVector3
{
public:
	inline BVector3(T x, T y, T z) { this->x = x; this->y = y; this->z = z; }
	BVector3() {}
	~BVector3() {}

	inline T Length() { return sqrt(x * x + y * y + z * z); }

	inline friend BVector3 operator+(const BVector3 v1, const BVector3 v2)
	{
		return BVector3<T>(v1.x + v2.x, v1.y + v2.y, v1.z + v2.z);
	}
	inline friend BVector3 operator-(const BVector3 v1, const BVector3 v2)
	{
		return BVector3<T>(v1.x - v2.x, v1.y - v2.y, v1.z - v2.z);
	}


	inline static BVector3 crossProduct(const BVector3 &v1, const BVector3 &v2)
	{
		return BVector3(
			v1.y * v2.z - v1.z * v2.y,
			v1.z * v2.x - v1.x * v2.z,
			v1.x * v2.y - v1.y * v2.x);
	}

	inline BVector3<float> toFloat()
	{
		return BVector3<float>((float)x, (float)y, (float)z);
	}

	T x, y, z;
};

// This class is similar to the boost matrix library and uses information from http://www.parashift.com/c++-faq-lite/const-correctness.html#faq-18.12

template<class T>
class BMatrix
{
public:
	BMatrix() : m(NULL) {}
	BMatrix(unsigned int r, unsigned int c)
	{
		this->r = r;
		this->c = c;
		m = new T[r * c];
	}
	BMatrix(const BMatrix &m)
	{
		// copy constructor
		r = m.r;
		c = m.c;
		this->m = new T[r * c];
		for (unsigned int i = 0; i < r; i++)
			for (unsigned int j = 0; j < c; j++)
				this->m[i * c + j] = m(i, j);
	}
	BMatrix& operator=(const BMatrix &mat)
	{
		r = mat.r;
		c = mat.c;
		if (m != NULL)
			delete[] m;
		m = new T[r * c];
		for (unsigned int i = 0; i < r * c; i++)
			m[i] = mat.m[i];
		return *this;
	}
	~BMatrix()
	{
		if (m != NULL)
			delete[] m;
	}
	inline void setSize(unsigned int r, unsigned int c)
	{
		this->r = r;
		this->c = c;
		if (m != NULL)
			delete[] m;
		m = new T[r * c];
	}
	inline T& operator()(int row, int col)
	{
		return m[row * c + col];
	}
	inline T operator()(int row, int col) const
	{
		return m[row * c + col];
	}

	inline static void prod(BMatrix<T> &res, const BMatrix<T> &m1, const BMatrix<T> &m2)
	{
		for (unsigned int i = 0; i < m1.r; i++)
			for (unsigned int j = 0; j < m2.c; j++)
			{
				T r = 0;
				for (unsigned int k = 0; k < m1.c; k++)
					r += m1(i, k) * m2(k, j);
				res(i, j) = r;
			}
	}

	inline BMatrix prod(const BMatrix &m2)
	{
		BMatrix res(r, m2.c);
		for (unsigned int i = 0; i < r; i++)
			for (unsigned int j = 0; j < m2.c; j++)
			{
				T r = 0;
				for (unsigned int k = 0; k < c; k++)
					r += m[i * c + k] * m2(k, j);
				res(i, j) = r;
			}
		return res;
	}

	inline BMatrix operator+=(const BMatrix &mat)
	{
		for (unsigned int i = 0; i < r * c; i++)
			m[i] += mat.m[i];
		return *this;
	}

	unsigned int rows() { return r; }
	unsigned int columns() { return c; }
private:
	T *m;
	unsigned int r;
	unsigned int c;
};

#endif
