#ifndef BOOTLOADER_H
#define BOOTLOADER_H

// wxWidgets includes
#include <wx/stattext.h>
#include <wx/msgdlg.h>
#include <wx/gauge.h>
#include <wx/notebook.h>

#include "bhand_misc.h"

// POCO library includes
#include "Poco/Thread.h"
#include "Poco/Runnable.h"
#include "Poco/Event.h"

using Poco::Thread;
using Poco::Runnable;
using Poco::Event;

enum UploadState
{
	Upload_Idle = 0,
	Upload_Wait,
	Upload_Uploading
};

class wxGauge;
class wxString;
class wxStaticText;
class wxMessageDialog;

class BHand;

class BGUIBootloader : public Runnable
{
public:
	BGUIBootloader(BHand *bhand, wxNotebook *bhNotebook, wxGauge* gauge, wxStaticText* firmOH, wxStaticText* helpMSG) :
		m_bhand(bhand), m_bhNotebook(bhNotebook), m_gauge(gauge), m_firmwareOH(firmOH), m_helpMSG(helpMSG),
		m_file_path(wxT("")),
		uploadState(Upload_Idle)
	{}
	virtual ~BGUIBootloader()
	{
		if (m_thread.isRunning())
		{
			m_closing.set();
			m_thread.join();
		}
	}

	// Method that starts the upload
	void UploadStart(int port)
	{
		// make sure a valid file exists
		if (!m_file_path.Cmp(wxT("")))
		{
			wxMessageDialog response(NULL, wxT("No firmware file selected.\nSelected path: ") + m_file_path, wxT("Invalid Firmware"), wxOK | wxICON_WARNING);
			response.ShowModal();
			return;
		}

		// prepare progress bar
		m_gauge->SetRange(1000);
		m_gauge->SetValue(0);


		// Do not start upload unless there is no upload in progress
		if (uploadState == Upload_Idle && UploadPrepareFirmware(port))
			uploadState = Upload_Wait;
		else
			return;

		m_bhNotebook->Disable(); // disable GUI while uploading
	}

	void UploadService()
	{
		switch (uploadState)
		{
			case Upload_Idle:
			{
				break;
			}
			case Upload_Wait:
			{
				if (UploadWaitReset())
					uploadState = Upload_Uploading;
				else
				{
					uploadState = Upload_Idle;
					UploadCleanup(true);
				}

				break;
			}
			case Upload_Uploading:
			{
				m_gauge->SetValue(m_gaugeValue);
				//m_gauge->Update();

				// multithreaded bootloader being used
				//UploadFirmwareLine();

				//if (m_thread.tryJoin(0)) // works under Windows but will block under Unbuntu and make App unresponsive
				if (!m_thread.isRunning())
					UploadCleanup();

				break;
			}
		}

	}

	// Pure virtual methods
	virtual bool UploadPrepareFirmware(unsigned int comport) = 0;
	virtual bool UploadWaitReset() = 0;
	virtual void UploadCleanup(bool errorFlag = false) = 0;

	virtual bool UploadFirmwareLine() = 0;

	// Public helper methods
	void setUploadState(UploadState us) { uploadState = us; }
	UploadState getUploadState() { return uploadState; }

	void setFilePath(wxString path) { m_file_path = path; }


protected:
	BHand *m_bhand;
	wxNotebook *m_bhNotebook;
	wxGauge *m_gauge;
	wxStaticText *m_firmwareOH;
	wxStaticText *m_helpMSG;

	wxString m_file_path;

	UploadState uploadState;

	// Multi-threading variables
	Thread	m_thread;
	Event m_closing;

	volatile int m_gaugeValue;

	///////////////////////////////////////////////////////////////////////////
	// Private method (required to be implemented to be Runnable)
	///////////////////////////////////////////////////////////////////////////

	void run()
	{
		//printf("run()\n");
		while (!UploadFirmwareLine())
		{
			//printf("here ");
			if (m_closing.tryWait(0))
				break;
		}
		//printf("run() done\n");
	}

};

#endif
