#include "Bootloader262.h"

#ifdef BH8_262_HARDWARE


#include <wx/file.h>

#include "BHand.h"
#include "BHandSerialDriver.h"


// for firmware uploading
static FILE *fhook;
static int lobyte, hibyte;
static int ramarray[300];
static char stype[3],shex[80], line[100];
static int num,opcode,temp_lo,temp_high,first;
static int total_bytes = 0, count = 0;
static int mycounter = 1;


Bootloader262::Bootloader262(BHand *bhand, wxNotebook *bhNotebook, wxGauge* gauge, wxStaticText* firmOH, wxStaticText* helpMSG) :
	BGUIBootloader(bhand, bhNotebook, gauge, firmOH, helpMSG)
{

}

Bootloader262::~Bootloader262()
{
}

///////////////////////////////////////////////////////////////////////////////
// Create 3 helper functions dealing with serial I/O
///////////////////////////////////////////////////////////////////////////////

char Bootloader262::Read()
{
	// Read character, check for errors
	char buf;
	return m_bhand->ComRead(&buf, 1) ? '\0' : buf;
}

void Bootloader262::Write(char ch)
{
	m_bhand->ComWrite(&ch, 1);
}

void Bootloader262::EchoWrite(char ch)
{
	// Echo read
	Write(ch);
	if (ch != Read())
		return;
}

bool Bootloader262::UploadPrepareFirmware(unsigned int comport)
{
	// prepare firmware uploading to the hand

	// Stop communication with the hand
	if (m_bhand->ComIsOpen())
	{
		printf("closing com port\n");
		m_bhand->ComClose();
	}


	printf("Bootloader262::UploadPrepareFirmware(%d)\n", comport);

	// Open serial port communication with hand
	if (m_bhand->ComOpen(comport))
	{
		wxMessageDialog response(NULL, wxT("Unable to open com port."), wxT("Serial Port Problem"), wxOK | wxICON_WARNING);
		response.ShowModal();
		return false;
	}

	// open a file stream to the firmware file
	fhook = fopen(m_file_path.mb_str(), "r");

	total_bytes = 0;

	while (!feof(fhook))
	{
		fgets(line,85,fhook);

		sscanf(line,"%2s%2x%2x%2x%76s", stype, &num, &temp_high, &temp_lo, shex);

		if (strstr(stype, "S1"))
		{
			num -= 3; // subtract addr. and checksum from # of pairs
			total_bytes += num;
		}
	}
	fclose(fhook);

	//printf("Pre-scanned firmware... \n");


	wxString message = wxT("Please power off the hand then click the \"OK\" Button.");
	wxMessageDialog notice(NULL, message, wxT("Info"), wxOK);
	notice.ShowModal();

	//printf("Cycle power... \n");

	// NOTE: when the hand is turned on, it first sends a ':' character.  After this character is sent,
	// there is a small window (250,000 ms approx) in which the ((char)(255)) can be sent to the hand to issue
	// capabilities for a firmware update.  The hand has to be power cycled to correctly perform a firmware update.
	// It is possible to update the firmware with a call to m_hand->Reset() and not power cycle, but it is less
	// reliable and slightly more complicated.

	// wait for the hand to be turned on
	m_helpMSG->Show(true);

	return true;
}

bool Bootloader262::UploadWaitReset()
{
	// waits until the hand has been power cycled and then readies hand for firmware uploading
	int errcnt = 0;

	while (Read() != ':')
	{
		if (++errcnt >= 2000)
		{
			wxString message = wxT("Download Failed - Error");
			wxMessageDialog notice(NULL, message, wxT("Error"), wxOK);
			notice.ShowModal();
			uploadState = Upload_Idle;
			return false;
		}

		DELAY(10);
	}

	m_helpMSG->Show(false);

	//printf("Past initial error check... \n");

	// send character 255 to the hand
	Write((char)255); // send first trigger byte

	// open the file stream and prepare to start writing firmware to the hand
	fhook = fopen(m_file_path.mb_str(), "r");
	strcpy(stype, "");
	first = 1;

	//printf("Opened file again... \n");

	mycounter = 1;
	count = 0;

	m_gaugeValue = 0;

	// Create thread for uploading firmware
	m_thread.setName("Upload firmware thread");
	m_thread.setPriority(Poco::Thread::PRIO_HIGHEST);
	m_thread.start(*this);

	return true;
}


void Bootloader262::UploadCleanup(bool errorFlag)
{
	// cleanup streams / initialize software after uploading

	// Close the file stream and return
	fclose(fhook);

	// Close the serial communication with hand
	m_bhand->ComClose();

	// Clear progress bar
	m_gauge->SetValue(0);
	m_gauge->Update();

	// Clear firmware label (it will display updated version after the hand connects)
	m_firmwareOH->SetLabel(wxT(""));

	m_bhNotebook->Enable();

	// Return upload state to the idle state
	uploadState = Upload_Idle;
}

bool Bootloader262::UploadFirmwareLine()
{
	// parses a line of firmware from the file (a file stream is kept open between calls)
	// and sends it to the hand

	if (!feof(fhook))
	{
		fgets(line, 85, fhook);
		if (feof(fhook))
			return true;

		// some C black magic
		sscanf(line,"%2s%2x%2x%2x%76s", stype, &num, &temp_high, &temp_lo, shex); // basic line format
		lobyte = temp_lo;
		hibyte = temp_high;
		if (strstr(stype, "S1"))
		{
			num -= 3; // subtract addr. and checksum from # of pairs
			float per = (100 * (double)(count + num) / (double)total_bytes) - 1;
			if (per < 0.0) per = 0.0;

			count += num;
			m_gaugeValue = (int)(per * 10.0);

			for (int i = 0; i < num; i++)
			{
				// read pairs into ramarray
				sscanf(shex + i * 2, "%2x", &opcode);
				ramarray[i] = opcode;
			}

			// write through the serial port

			Write((char)255);

			EchoWrite(lobyte);
			EchoWrite(hibyte);
			EchoWrite(num);
			Write((char)255);
			for (int i = 0; i < num; i++)
				EchoWrite(ramarray[i]);
		}

		//printf("Processed line %d\n", mycounter);
		mycounter++;
	}
	else
	{
		return true;
	}

	return false;
}

#endif
