#include "Config.h"

#include <wx/wx.h>
#include <wx/gauge.h>
#include <wx/stattext.h>
#include <wx/file.h>
#include <wx/bitmap.h>

#include "Common/mathplot.h"
#include "Common/tmr.h"

#include "Bootloader/Bootloader262.h"
#include "Bootloader/BootloaderPuck2.h"

#define NUM_DELAY_TESTS 100

extern bool use280Config;

Config::Config(wxNotebook *controlNotebook, BHand* bh, wxChoice *comportChoice, wxGauge* gauge, wxStaticText* max, wxStaticText* min, wxStaticText* avg, wxStaticText* outin, wxStaticText* firmOH, wxStaticText* helpMSG) :
	// Set member variables from those that were passed to the constructor
	bhNotebook(controlNotebook), m_hand(bh), m_bhcontrol(NULL), m_comportChoice(comportChoice),

	m_max(max), m_min(min), m_avg(avg), m_outin(outin),

	m_firmwareOH(firmOH),

	// Init other member variables
	m_use_CAN(true), initLibraryState(BHInitLibraryIdle)
{
	// Setup bootloader(s)
#ifdef BH8_262_HARDWARE
	m_bootloader262 = new Bootloader262(m_hand, bhNotebook, gauge, firmOH, helpMSG);
#endif
#ifdef BH8_280_HARDWARE
	m_bootloaderPuck2 = new BootloaderPuck2(m_hand, bhNotebook, gauge, firmOH, helpMSG);
#endif

	m_bootloader = NULL;

	m_initialized = false;

	// Get a local copy of the pointer to the config tab
	m_panel = BHControlFrame::getBHControlFrame()->Configure_Tab;


	// loads an image for the bar graphs for test delays
	if (wxFile::Exists((wxT("src/Common/bar.bmp"))))
		box.LoadFile(wxT("src/Common/bar.bmp"));
	else if (wxFile::Exists((wxT("../src/Common/bar.bmp"))))
		box.LoadFile(wxT("../src/Common/bar.bmp"));
	else if (wxFile::Exists((wxT("../../src/Common/bar.bmp"))))
		box.LoadFile(wxT("../../src/Common/bar.bmp"));

	// Initialize graph widget for the Configuration Tab
	m_histogram = new mpWindow(m_panel, -1, wxPoint(296,320), wxSize(442,298), wxSUNKEN_BORDER);

	// test delays axes are milliseconds and number (count)
	mpScaleX* xaxis = new mpScaleX(wxT("X (ms)"), mpALIGN_BOTTOM, true, mpX_NORMAL);
	mpScaleY* yaxis = new mpScaleY(wxT("Y (count)"), mpALIGN_LEFT, true);

	m_histogram->AddLayer(xaxis);
	m_histogram->AddLayer(yaxis);

	xaxis->SetDrawOutsideMargins(false);
	yaxis->SetDrawOutsideMargins(false);

	m_histogram->SetScaleX(10.0);
	m_histogram->SetScaleY(10.0);
	m_histogram->SetScr(10,10);

	// NOTE: needed on windows, include it anyway on linux
	m_histogram->EnableDoubleBuffer(true);

	m_histogram->SetMargins(0, 0, 20, 40);
	m_histogram->LockAspect(false);

	// initialize graph for test delays
	for (int i = 0; i < NUM_DELAY_TESTS; i++)
	{
		bars.push_back(new mpBitmapLayer());
		m_histogram->AddLayer(bars[i]);
	}
}

Config::~Config()
{
    m_histogram->DelAllLayers(true, false);
    bars.clear();

	delete m_histogram;

#ifdef BH8_262_HARDWARE
	delete m_bootloader262;
#endif
#ifdef BH8_280_HARDWARE
	delete m_bootloaderPuck2;
#endif
}


void Config::setBHControlFrame(BHControlFrame *bhcontrol)
{
	m_bhcontrol = bhcontrol;
}


void Config::setBaudRate(int baudrate)
{
	if (!m_hand->Baud(baudrate))
	{
		wxString message = wxT("Baud Rate set to: ") + wxString::Format(wxT("%i"),baudrate) +
			wxT("\nExecute a PSAVE command in the terminal Window to save global parameters") +
			wxT("\nto set baud rate permanently.  Examples and demos run only at 9600 bps.");
		wxMessageDialog response (NULL, message, wxT("Info"), wxOK);
		response.ShowModal();
	}
	else
		ThrowError(wxT("Baud Rate has not been set."));
}


///////////////////////////////////////////////////////////////////////////////
// Events called by BHControlMain
///////////////////////////////////////////////////////////////////////////////

void Config::OnBaudRate(wxCommandEvent& event)
{
	if (m_initialized)
	{
		long baudrate;

		if ((m_bhcontrol->CONFIG_BaudRate_Choice->GetStringSelection()).ToLong(&baudrate))
			setBaudRate(baudrate);
	}
}


void Config::OnSetConfigDefaults(wxCommandEvent& event)
{
	// re-initializes hand to default settings (does not re-initialize software)
	if (!m_use_CAN)
	{
		m_bhcontrol->CONFIG_BaudRate_Choice->SetSelection(0);
		setBaudRate(9600);
	}
	ResetDefaults();
}


void Config::setFilePath(wxString path)
{
	if (m_bootloader != NULL)
		m_bootloader->setFilePath(path);
}


void Config::setUseCAN(bool value)
{
	m_use_CAN = value;
#ifdef BH8_262_HARDWARE
	if (!m_use_CAN)
		m_bootloader = m_bootloader262;
#endif
#ifdef BH8_280_HARDWARE
	if (m_use_CAN)
		m_bootloader = m_bootloaderPuck2;
#endif
}


bool Config::initialize(int hwIndex)
{
	if (initLibraryState == BHInitLibraryIdle)
	{
		// Set hardware description (list with available properties)
		m_hand->setHardwareDesc(hwIndex);
		use280Config = (strcmp(m_hand->getHardwareDesc()->getModelNumber(), "BH8-280") == 0);

		initLibraryState = BHInitLibraryStart;
		return true;
	}

	return false;
}

void Config::ResetDefaults()
{
	if (m_hand->Reset())
	{
		ThrowError(wxT("Error Resetting Hand."));
		return;
	}

	if (m_bhcontrol->CONFIG_AutoInit_CheckBox->GetValue())
	{
		if (m_hand->InitHand(""))
		{
			ThrowError(wxT("Error Resetting Hand."));
			return;
		}
	}
}


void Config::ThrowError(wxString message)
{
    wxMessageDialog response(NULL, message, wxT("Configure - Error"), wxOK | wxICON_ERROR);
    response.ShowModal();
}

int compare(const void *arg1, const void *arg2)
{
	// compare function used in quicksort for TestDelays()
	if (*(unsigned int*)arg1 < *(unsigned int*)arg2)
		return -1;
	else if (*(unsigned int*)arg1 == *(unsigned int*)arg2)
		return 0;
	else
		return 1;
}

void Config::OnTestDelays(wxCommandEvent& event)
{
	// Create motor string
    int current = 0;

	char motors[5];
    if (m_bhcontrol->CONFIG_Motor1_CheckBox->GetValue()) motors[current++] = '1';
    if (m_bhcontrol->CONFIG_Motor2_CheckBox->GetValue()) motors[current++] = '2';
    if (m_bhcontrol->CONFIG_Motor3_CheckBox->GetValue()) motors[current++] = '3';
    if (m_bhcontrol->CONFIG_Motor4_CheckBox->GetValue()) motors[current++] = 'S';

    motors[current] = '\0';

	// Get motor control flags
	bool control_velocity_flag = m_bhcontrol->CONFIG_VelocityControl_CheckBox->GetValue();
	bool control_propgain_flag = m_bhcontrol->CONFIG_PropGainControl_CheckBox->GetValue();

	// Get motor feedback flags from checkboxes
	bool feedback_velocity_flag = m_bhcontrol->CONFIG_VelocityFeedback_CheckBox->GetValue();
	bool feedback_strain_flag = m_bhcontrol->CONFIG_StrainFeedback_CheckBox->GetValue();
	bool feedback_position_flag = m_bhcontrol->CONFIG_PosFeedback_CheckBox->GetValue();
	bool feedback_deltapos_flag = m_bhcontrol->CONFIG_DpFeedback_CheckBox->GetValue();

	// Set control and feedback boolean parameters for RTUpdate
	bool control = control_velocity_flag || control_propgain_flag;
	bool feedback = feedback_velocity_flag || feedback_strain_flag ||
	                feedback_position_flag || feedback_deltapos_flag;

    // set other variables used for RTSetFlags
    int control_velocity_coefficient = 1;
    int feedback_velocity_coefficient = 1;
    int feedback_delta_position_coefficient = 1;

	if (use280Config)
		m_hand->Set("", "LCP", 0);
	m_hand->Set("", "LCT", 0);

	int result;
	if ((result = m_hand->RTSetFlags(motors,
                                  control_velocity_flag, control_velocity_coefficient, control_propgain_flag,
                                  feedback_velocity_flag, feedback_velocity_coefficient, feedback_strain_flag,
                                  feedback_position_flag, feedback_deltapos_flag, feedback_delta_position_coefficient)))
	{
		ThrowError(wxT("Set RT Flags failed."));
        return;
    }

	// start RT mode
	if ((result = m_hand->RTStart(motors)))
	{
		ThrowError(wxT("Start RT mode failed."));
        return;
	}

	// set control parameters
	for (int m = 0; m < 4; m++)
	{
		if (control_velocity_flag)
			m_hand->RTSetVelocity('1' + m, 0);
		if (control_propgain_flag)
			m_hand->RTSetGain('1' + m, 8);
	}

	// test delays
	//printf("TestDelays() calling RTUpdate(%d, %d)\n", control, feedback);
	float isi[NUM_DELAY_TESTS];

    HiResTimer t;

	for (int i = 0; i < NUM_DELAY_TESTS; i++)
	{
		t.Start();
		if ((result = m_hand->RTUpdate(control, feedback)))
		{
			ThrowError(wxT("RTUpdate call failed to update in real-time."));
            return;
        }
        t.Stop();

        isi[i] = (float)t.GetDurationInMilliSecs();

		// check for long delays
		if (isi[i] > 950)
		{
			m_hand->RTAbort();
			ThrowError(wxT("Warning - Delay was too long"));
			return;
		}
	}


	// end RT mode
	if ((result = m_hand->RTAbort()))
		ThrowError(wxT("RTAbort call failed"));


	// sort, compute statistics
	qsort((void*)isi, NUM_DELAY_TESTS, sizeof(float), compare);

    //for (int i = 0; i < NUM_DELAY_TESTS; i++)
    //    printf("val %d: %f\n", i, isi[i]);


    // fill in test delay variables such as min, max, avg, out/in, etc

	// Min
	m_min->SetLabel(wxString::Format(wxT("%i"), (int)isi[0]));

	// Max
	m_max->SetLabel(wxString::Format(wxT("%i"), (int)isi[NUM_DELAY_TESTS-1]));

	// Avg
	float sum = 0;
	for (int i = 0; i < NUM_DELAY_TESTS; i++ )
		sum += isi[i];
	sum /= NUM_DELAY_TESTS;

	m_avg->SetLabel(wxString::Format(wxT("%.2f"), sum));

	// Print results to console
	//printf("min= %d ms, max = %d ms, avg = %1.0f ms\n", (int)isi[0], (int)isi[NUM_DELAY_TESTS-1], sum);
    //printf("Done!\n");

    //use delays in isi array to make a histogram
	DrawHistrogram(isi, NUM_DELAY_TESTS);
}


void Config::DrawHistrogram(float* array, int size)
{
	// draw a histogram displaying information about delay times in RT mode

	//set the remaining bars to zero
	for (int i = 0; i < size; i++)
		bars[i]->SetVisible(false);

	//create a placeholder array and set all values to zero
	int delay_times_size = (int)(array[size-1]) + 1;
	int *delay_times = new int[delay_times_size];

	for (int i = 0; i < (int)(array[size-1] + 1); i++)
		delay_times[i] = 0;

	int temp;

	//there are 100 results in the float* array
	//increment values of delay_times by one for each corresponding value in the float* array

	for(int i = 0; i < size; i++)
	{
		temp = (int)array[i];
		delay_times[temp]++;
	}

	//for(int i = 0; i < (int)(array[size-1] + 1); i++)
	//	printf("number of runs at delay %i: %i\n", i, delay_times[i]);

	int counter = 0;

	//set bars according to delay_times
	//only increment up to the highest value in the array
	for (int i = 0; i < (int)array[size-1] + 1; i++)
	{
		if(delay_times[i] > 0)
		{
			bars[counter]->SetBitmap(box, ((float)i)-0.2, 0, 0.4, delay_times[i]);
			bars[counter]->SetVisible(true);
			counter++;
		}
	}

	//display the bar graph
	m_histogram->Fit();

	delete[] delay_times;
}


///////////////////////////////////////////////////////////////////////////////
// GUI Methods for Bootloaders
///////////////////////////////////////////////////////////////////////////////

void Config::UploadStart()
{
#ifdef LINUX
	if (m_bootloader != NULL)
		m_bootloader->UploadStart(m_comportChoice->GetSelection() - 9);
#else
	if (m_bootloader != NULL)
		m_bootloader->UploadStart(m_comportChoice->GetSelection() + 1);
#endif
}


void Config::UploadService()
{
	if (m_bootloader != NULL)
		m_bootloader->UploadService();

	wxString message;
	static int vers[4] = { 0, 0, 0, 0 };
	static int poles[4] = { 0, 0, 0, 0 };
	static char getVersText[5000] = "";

	// Handle "Connect Library" event mostly aysnchronous
	switch (initLibraryState)
	{
	case BHInitLibraryIdle:
		{
			break;
		}
	case BHInitLibraryStart:
		{
			///////////////////////////////////////////////////////////////////////////
			// Intialize the software
			///////////////////////////////////////////////////////////////////////////

			if (m_use_CAN)
			{
				// CAN Interface
				m_hand->Init(0, THREAD_PRIORITY_TIME_CRITICAL, BH_CAN_COMMUNICATION, true); // error codes checked in next state
			}
			else
			{
				long baudrate;
#ifdef LINUX
				int com_port = m_bhcontrol->CONFIG_Com_Choice->GetSelection() - 9;
#else
				int com_port = m_bhcontrol->CONFIG_Com_Choice->GetSelection() + 1;
#endif

				if ((m_bhcontrol->CONFIG_BaudRate_Choice->GetStringSelection()).ToLong(&baudrate))
					BHand::SetDefaultBaud(baudrate);

				// Serial Interface
				m_hand->Init(com_port, THREAD_PRIORITY_TIME_CRITICAL, BH_SERIAL_COMMUNICATION, true); // error codes checked in next state
			}

			initLibraryState = BHInitLibraryWaitForReset;

			break;
		}
	case BHInitLibraryWaitForReset:
		{
			if (m_hand->ComIsPending())
				break;

			if (!m_use_CAN)
				m_hand->ComSetTimeouts(50, 15000, 50, 5000);

			// Reset complete - check the error code
			int result;
			if ((result = m_hand->ComGetError()))
			{
				// Error with Init software call
				if (m_use_CAN)
				{
					message = wxString::Format(wxT("Error Initializing Software. Check connections. %d"), result);
					ThrowError(message);
				}
				else
				{
					message = wxString::Format(wxT("Error Initializing Software. Make sure Com Port is correct. %d"), result);
					ThrowError(message);
				}
				initLibraryState = BHInitLibraryIdle;
				break;
			}

			// Request version
			if (m_use_CAN)
				m_hand->Get("1234", "VERS", vers);
			else
				m_hand->Command("VERS");

			initLibraryState = BHInitLibraryWaitForVersResult;

			break;
		}
	case BHInitLibraryWaitForVersResult: // CAN only
		{
			if (m_hand->ComIsPending())
				break;

			int result;
			if ((result = m_hand->ComGetError()))
			{
				// Error getting vers property
				message = wxString::Format(wxT("Error getting VERS property. Check connections. %d"), result);
				ThrowError(message);

				initLibraryState = BHInitLibraryIdle;
				break;
			}

			if (m_use_CAN)
			{
				// Get actual Vers result
				BHandResultGet *r = (BHandResultGet*)m_hand->GetResult();
				vers[0] = r->getResult(0);
				vers[1] = r->getResult(1);
				vers[2] = r->getResult(2);
				vers[3] = r->getResult(3);

				m_hand->Get("1234", "POLES", poles);
				initLibraryState = BHInitLibraryWaitForPolesResult;
			}
			else
			{
#ifdef BH8_262_HARDWARE
				strcpy(getVersText, m_hand->Response());
#endif

				// initialize motors (depends on checkbox)
				if (m_bhcontrol->CONFIG_AutoInit_CheckBox->GetValue())
				{
					m_hand->InitHand("");
					initLibraryState = BHInitLibraryWaitForInitHand;
				}
			}

			break;
		}
	case BHInitLibraryWaitForPolesResult: // CAN only
		{
			if (m_hand->ComIsPending())
				break;

			int result;
			if ((result = m_hand->ComGetError()))
			{
				// Error getting poles property
				message = wxString::Format(wxT("Error getting POLES property. Check connections. %d"), result);
				ThrowError(message);

				initLibraryState = BHInitLibraryIdle;
				break;
			}

			// Check actual Poles result
			BHandResultGet *r = (BHandResultGet*)m_hand->GetResult();
			poles[0] = r->getResult(0);
			poles[1] = r->getResult(1);
			poles[2] = r->getResult(2);
			poles[3] = r->getResult(3);

			// Check that Poles is not equal to 6
			bool resetDefaults = false;
			for (int i = 0; i < 4; i++)
			{
				if (vers[i] > 0 && poles[i] != 6)
					resetDefaults = true;
			}

			if (resetDefaults)
			{
				message = wxT("Motor poles property is not set to 6.\nThis may be caused by a shift in EEPROM after uploading new firmware.\nClick OK to reset Pucks in the hand to default values.");

				wxMessageDialog response(NULL, message, wxT("Invalid Motor Property Found"), wxOK | wxICON_WARNING);
				response.ShowModal();

				m_hand->Default("123S");
				while (m_hand->ComIsPending()); // Block until default command runs asynchronously

				message = wxT("Please restart the hand and the GUI.");

				wxMessageDialog response2(NULL, message, wxT("Default values set"), wxOK | wxICON_WARNING);
				response2.ShowModal();

				initLibraryState = BHInitLibraryIdle;
				break;
			}

			// initialize motors (depends on checkbox)
			if (m_bhcontrol->CONFIG_AutoInit_CheckBox->GetValue())
			{
				m_hand->InitHand("");
				initLibraryState = BHInitLibraryWaitForInitHand;
			}

			break;
		}
	case BHInitLibraryWaitForInitHand:
		{
			if (m_hand->ComIsPending())
				break;

			int result;
			if ((result = m_hand->ComGetError()))
			{
				// Error from InitHand() call
				message = wxString::Format(wxT("Error Initializing Hand. %d"), result);
				ThrowError(message);

				initLibraryState = BHInitLibraryIdle;
				break;
			}


			if (m_use_CAN)
			{
				m_firmwareOH->SetLabel(wxString::Format(wxT("%d, %d, %d, %d"), vers[0], vers[1], vers[2], vers[3]));
			}
			else
			{
#ifdef BH8_262_HARDWARE
				// check version
				wxString version_number(getVersText, wxConvUTF8);

				m_firmwareOH->SetLabel(version_number); // Updates firmware label

				// delay an info box if successful
				long baudrate; (m_bhcontrol->CONFIG_BaudRate_Choice->GetStringSelection()).ToLong(&baudrate);
#ifdef LINUX
				int com_port = m_bhcontrol->CONFIG_Com_Choice->GetSelection() - 9;//+ 1;
#else
				int com_port = m_bhcontrol->CONFIG_Com_Choice->GetSelection() + 1;
#endif
				message = wxT("Initialized Barrett Hand.\nFirmware: ") + version_number + wxT("\n") +
						  wxT("Baud Rate: ") +  wxString::Format(wxT("%i"), (int)baudrate) +
						  wxT("\nCom Port: ") + wxString::Format(wxT("%i"), com_port) + wxT("\n");


				version_number = version_number.SubString(8, 10);

				// make sure the firmware value is 4.xxx
				double value;
				if (!version_number.ToDouble(&value))
				{
					message += wxT("Failed to parse firmware version number, exiting.");
					wxMessageDialog response(NULL, message, wxT("Invalid Firmware"), wxOK | wxICON_WARNING);
					response.ShowModal();
					initLibraryState = BHInitLibraryIdle;
					break;
				}

				if (value < 4)
				{
					message = message + wxT("This software is designed for \nFirmware Version 4.4.0 or higher.\n\nPlease download a recent .S19 file.\n") +
										wxT("NOTE: wait at least 30 seconds before re-initializing in order to prevent damage to the Barrett Hand.");
					wxMessageDialog response(NULL, message, wxT("Invalid Firmware"), wxOK | wxICON_WARNING);
					response.ShowModal();
					initLibraryState = BHInitLibraryIdle;
					break;
				}

				wxMessageDialog response (NULL, message, wxT("Info"), wxOK);
				response.ShowModal();

#endif
			}

			// Finish initialization
			m_bhcontrol->CONFIG_Browse_Button->Disable();
			m_bhcontrol->CONFIG_Upload_Button->Disable();
			m_bhcontrol->handsChoice->Disable();
			m_bhcontrol->CONFIG_CAN_RadioButton->Disable();
			m_bhcontrol->CONFIG_Serial_RadioButton->Disable();
			m_bhcontrol->CONFIG_AutoInit_CheckBox->Disable();

			m_initialized = true;

			m_hand->setSyncMode(BHMODE_SYNC);
			m_bhcontrol->SetDefaults(true);
			initLibraryState = BHInitLibraryIdle;

			break;
		}

	};
}
