#include "BHandKinematics.h"

#include <math.h>
#include <stdio.h>

const double PI = 3.1415926535897932384626433832795;

#define DEG2RAD(x) (((x) * PI) / 180.0)

double BHandKinematics::dhAParam[]     = { 0.025,       0.050, 0.070,        0.050 }; // { Aw, A1, A2, A3 } units in mm (note Aw needs to be multiplied by r in code)
double BHandKinematics::dhAlphaParam[] = { 0, PI / 2.0,  0, -PI / 2.0 }; // { alphaw, alpha1, alpha2, alpha3 }
double BHandKinematics::dhDParam[]     = { 0.084,        0,  0,       0.095 }; // { Dw, D1, D2, D3 } units in mm
//const double dhDParam[]     = { 84,        0,  0,       9.5 }; // { Dw, D1, D2, D3 } units in mm (in manual)


void print(BMatrix<double> &m)
{
	for (unsigned int r = 0; r < m.rows(); r++)
	{
		for (unsigned int c = 0; c < m.columns(); c++)
			printf("%8.3f ", m(r, c));
		printf("\n");
	}
}


BHandKinematics::BHandKinematics() :
	capTheta(4, 4)
{
	//capPhi2 = DEG2RAD(2.46); // 307.5 BH8-262 encoder ticks (value in manual)
	capPhi2 = DEG2RAD(0); // 307.5 BH8-262 encoder ticks
	capPhi3 = DEG2RAD(42); // adjusted value
	//capPhi3 = DEG2RAD(50.0); (value in manual)
}

BHandKinematics::~BHandKinematics()
{
}

void BHandKinematics::fillCapThetaBefore(double *pos)
{
	// before torque switch activation
	capTheta(1, 2) = DEG2RAD((1.0/125.0) * pos[0]);
	capTheta(1, 3) = DEG2RAD((1.0/375.0) * pos[0]);
	capTheta(2, 2) = DEG2RAD((1.0/125.0) * pos[1]);
	capTheta(2, 3) = DEG2RAD((1.0/375.0) * pos[1]);
	capTheta(3, 2) = DEG2RAD((1.0/125.0) * pos[2]);
	capTheta(3, 3) = DEG2RAD((1.0/375.0) * pos[2]);
	capTheta(1, 1) = DEG2RAD((2.0/35.0) * pos[3]); // changed signed to be positive
	capTheta(2, 1) = DEG2RAD((2.0/35.0) * pos[3]);
	capTheta(3, 1) = 0;
}

void BHandKinematics::fillCapTheta(double *motorEncPos, double *innerEncPos)
{
	// this method takes encoder positions with units of Radians
	capTheta(1, 2) = motorEncPos[0]; // F1
	capTheta(1, 3) = (innerEncPos != NULL) ? innerEncPos[0] : (motorEncPos[0] / 3); // assume no breakaway if innerEncPos is NULL

	capTheta(2, 2) = motorEncPos[1]; // F2
	capTheta(2, 3) = (innerEncPos != NULL) ? innerEncPos[1] : (motorEncPos[1] / 3);

	capTheta(3, 2) = motorEncPos[2]; // F3
	capTheta(3, 3) = (innerEncPos != NULL) ? innerEncPos[2] : (motorEncPos[2] / 3);

	capTheta(1, 1) = motorEncPos[3]; // Spread
	capTheta(2, 1) = motorEncPos[3];
	capTheta(3, 1) = 0;
}

void BHandKinematics::fillF1MatrixWtoT(BMatrix<double> &m)
{
	int k = 1; // fill matrix using Forward Kinematics transformation matrix for finger 1
	double r = -1.0; // hardcoded for finger 1
	double j = 1.0; // hardcoded for finger 1

	double a = capTheta(k, 2) + capPhi2;
	double b = capTheta(k, 3) + capPhi3;

	double cos_a = cos(a);
	//double cos_b = cos(b);

	double sin_a = sin(a);

	double cos_ab = cos(a + b);
	double sin_ab = sin(a + b);
	double cos_1 = cos(r * capTheta(k, 1) - (PI/2.0) * j);
	double sin_1 = sin(r * capTheta(k, 1) - (PI/2.0) * j);

	double AW = dhAParam[0] * -1; // note that -1 is used for finger 1
	double A1 = dhAParam[1];
	double A2 = dhAParam[2];
	double A3 = dhAParam[3];

	double DW = dhDParam[0];
	//double D1 = dhDParam[1];
	//double D2 = dhDParam[2];
	double D3 = dhDParam[3];

	m(0, 0) = cos_1 * cos_ab;
	m(0, 1) = -sin_1;
	m(0, 2) = -cos_1 * sin_ab;
	m(0, 3) = (A3 * cos_1 * cos_ab) - (D3 * cos_1 * sin_ab) + (A2 * cos_1 * cos_a) + (A1 * cos_1) + (j * AW) ; // cos_a correct?

	m(1, 0) = sin_1 * cos_ab;
	m(1, 1) = cos_1;
	m(1, 2) = -sin_1 * sin_ab;
	m(1, 3) = (A3 * sin_1 * cos_ab) - (D3 * sin_1 * sin_ab) + (A2 * sin_1 * cos_a) + (A1 * sin_1);

	m(2, 0) = sin_ab;
	m(2, 1) = 0;
	m(2, 2) = cos_ab;
	m(2, 3) = (A3 * sin_ab) + (D3 * cos_ab) + (sin_a * A2) + DW;

	m(3, 0) = 0;
	m(3, 1) = 0;
	m(3, 2) = 0;
	m(3, 3) = 1;
}

void BHandKinematics::fillmatrix(int i, BMatrix<double> &m, double theta, double r)
{
	double theta_i = theta;
	double alpha_iminus1 = dhAlphaParam[i-1];
	double d_i = dhDParam[i-1];
	double a_iminus1 = dhAParam[i-1]; if (i == 1) a_iminus1 *= r;

	double cos_theta_i = cos(theta_i);
	double sin_theta_i = sin(theta_i);
	double cos_alpha_iminus1 = cos(alpha_iminus1);
	double sin_alpha_iminus1 = sin(alpha_iminus1);

	m(0, 0) = cos_theta_i;
	m(0, 1) = -sin_theta_i;
	m(0, 2) = 0;
	m(0, 3) = a_iminus1;

	m(1, 0) = sin_theta_i * cos_alpha_iminus1;
	m(1, 1) = cos_theta_i * cos_alpha_iminus1;
	m(1, 2) = -sin_alpha_iminus1;
	m(1, 3) = -sin_alpha_iminus1 * d_i;

	m(2, 0) = sin_theta_i * sin_alpha_iminus1;
	m(2, 1) = cos_theta_i * sin_alpha_iminus1;
	m(2, 2) = cos_alpha_iminus1;
	m(2, 3) = cos_alpha_iminus1 * d_i;

	m(3, 0) = 0;
	m(3, 1) = 0;
	m(3, 2) = 0;
	m(3, 3) = 1;
}

void BHandKinematics::getBHandMatrices(int k, BMatrix<double> *wTo1, BMatrix<double> *wTo2, BMatrix<double> *wTo3, BMatrix<double> *wToT) //BMatrix<double> **matrices, int numMatrices)
{
	double r, j;

    if (k == 1) { r = -1; j = 1; }
	else if (k == 2) { r = 1; j = 1; }
	else if (k == 3) { r = 0; j = -1; }
	else { k = 1; r = -1; j = 1; }

	if (wTo1 != NULL)
	{
		fillmatrix(1, *wTo1, r * capTheta(k, 1) - (PI / 2.0) * j, r); // removed r from the kinematic equations
		if (wTo2 != NULL)
		{
			BMatrix<double> matrix1to2(4, 4); fillmatrix(2, matrix1to2, capTheta(k, 2) + capPhi2, r);
			BMatrix<double>::prod(*wTo2, *wTo1, matrix1to2);
			if (wTo3 != NULL)
			{
				BMatrix<double> matrix2to3(4, 4); fillmatrix(3, matrix2to3, capTheta(k, 3) + capPhi3, r);
				BMatrix<double>::prod(*wTo3, *wTo2, matrix2to3);
				if (wToT != NULL)
				{
					BMatrix<double> matrix3toT(4, 4); fillmatrix(4, matrix3toT, 0, r);
					BMatrix<double>::prod(*wToT, *wTo3, matrix3toT);
				}
			}
		}
	}
}

/*void BHandKinematics::getBHandMatrices(int k, BMatrix<double> **matrices, int numMatrices)
{
	double r, j;

    if (k == 1) { r = -1; j = 1; }
	else if (k == 2) { r = 1; j = 1; }
	else if (k == 3) { r = 0; j = -1; }
	else { k = 1; r = -1; j = 1; }

	BMatrix<double> matrixWto1(4, 4); fillmatrix(1, matrixWto1, r * capTheta(k, 1) - (PI / 2.0) * j, r); // removed r from the kinematic equations
	BMatrix<double> matrix1to2(4, 4); fillmatrix(2, matrix1to2,     capTheta(k, 2) + capPhi2, r);
	BMatrix<double> matrix2to3(4, 4); fillmatrix(3, matrix2to3,     capTheta(k, 3) + capPhi3, r);
	BMatrix<double> matrix3toT(4, 4); fillmatrix(4, matrix3toT,     0, r);


	if (numMatrices >= 1)
		*(matrices[0]) = matrixWto1;
	if (numMatrices >= 2)
		*(matrices[1]) = matrixWto1.prod(matrix1to2);
	if (numMatrices >= 3)
		*(matrices[2]) = matrices[1]->prod(matrix2to3);
	if (numMatrices >= 4)
		*(matrices[3]) = matrices[2]->prod(matrix3toT);
}*/

void BHandKinematics::test()
{
	// void BHandKinematics::transformWtoT(int f, double *pos)
	double encTestPos1[] = { 0, 0, 0, 0 };
	fillCapThetaBefore(encTestPos1);

	///////////////////////////////////////////////////////////////////////////
	// Test wrist to tool matrix
	///////////////////////////////////////////////////////////////////////////

	BMatrix<double> f1MatrixWtoTDirect(4, 4);

	fillF1MatrixWtoT(f1MatrixWtoTDirect);

	///////////////////////////////////////////////////////////////////////////
	// Multiply out and solve for wrist to tool matrix
	///////////////////////////////////////////////////////////////////////////

	BMatrix<double> f1MatrixWtoTIndirect(4, 4);

	int k = 1; // fill matrix using Forward Kinematics transformation matrix for finger 1
	double r = -1.0; // hardcoded for finger 1
	double j = 1.0; // hardcoded for finger 1
	BMatrix<double> f1MatrixWto1(4, 4); fillmatrix(1, f1MatrixWto1, r * capTheta(k, 1) - (PI / 2.0) * j, -1);
	BMatrix<double> f1Matrix1to2(4, 4); fillmatrix(2, f1Matrix1to2,     capTheta(k, 2) + capPhi2,        -1);
	BMatrix<double> f1Matrix2to3(4, 4); fillmatrix(3, f1Matrix2to3,     capTheta(k, 3) + capPhi3,        -1);
	BMatrix<double> f1Matrix3toT(4, 4); fillmatrix(4, f1Matrix3toT,     0,                               -1);

	BMatrix<double> f1MatrixWto2(4, 4); f1MatrixWto2 = f1MatrixWto1.prod(f1Matrix1to2);
	BMatrix<double> f1MatrixWto3(4, 4); f1MatrixWto3 = f1MatrixWto2.prod(f1Matrix2to3);
	BMatrix<double> f1MatrixWto4(4, 4); f1MatrixWto4 = f1MatrixWto3.prod(f1Matrix3toT);

	// Printout Test
	print(f1MatrixWtoTDirect); printf("\n\n");

	print(f1MatrixWto1); printf("\n");
	print(f1Matrix1to2); printf("\n");
	print(f1Matrix2to3); printf("\n");
	print(f1Matrix3toT); printf("\n\n");

	print(f1MatrixWto2); printf("\n");
	print(f1MatrixWto3); printf("\n");
	print(f1MatrixWto4);

	//print(f1Matrix3toT); printf("\n");

}
