#include "HandManager.h"

#include "Models/HandBase.h"
#include "Models/HandJoint.h"
#include "Models/HandKnuckle.h"
#include "Models/HandMiddle.h"
#include "Models/HandFinger.h"
#include "Models/HandFingerSpreadable.h"

#include "Models/BoundingSphere.h"

#include "VisualControl.h"
#include "VisualFingerTip.h"
#include "VisualPalm.h"
#include "VisualData.h"

#include <math.h>

#define COLLISION_DEBUG false

#define PI 3.1415926535897932384626433832795

#define HAND_KINEMATICS_RADIUS 0.0095
//#define HAND_KINEMATICS_RADIUS 0.16

#define DEG2RAD(x) (((x) * PI) / 180.0)


class Color
{
public:
	Color() { c[0] = 0; c[1] = 0; c[2] = 0; }
	Color(float r, float g, float b)
	{
		c[0] = r;
		c[1] = g;
		c[2] = b;
	}
	~Color() {}

	float c[3];
};


class Material
{
public:
	Material(const Color &ambient, const Color &diffuse, const Color &specular, float shininess)
	{
		setAmbient(ambient);
		setDiffuse(diffuse);
		setSpecular(specular, shininess);
	}
	~Material() {}
	void setAmbient(const Color &c) { ambient = c; amb[0] = c.c[0]; amb[1] = c.c[1]; amb[2] = c.c[2]; amb[3] = 1; }
	void setDiffuse(const Color &c) { diffuse = c; dif[0] = c.c[0]; dif[1] = c.c[1]; dif[2] = c.c[2]; dif[3] = 1; }
	void setSpecular(const Color &c, float shininess) { specular = c; this->shininess = shininess; spec[0] = c.c[0]; spec[1] = c.c[1]; spec[2] = c.c[2]; spec[3] = 1; }
	const float *getAmbient() { return amb; }
	const float *getDiffuse() { return dif; }
	const float *getSpecular() { return spec; }
	float getShininess() { return shininess; }
private:
	float amb[4];
	float dif[4];
	float spec[4];
	Color ambient, diffuse, specular;
	float shininess;
};

void setMaterial(Material *m)
{
	GLfloat zero[] = { 0.0f, 0.0f, 0.0f, 1.0f };
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, m->getAmbient());
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, m->getDiffuse());
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, m->getSpecular());
	glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, zero);
	glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, m->getShininess());
}

Material jointMaterial(Color(0.24f, 0.20f, 0.08f), Color(0.75f, 0.61f, 0.23f), Color(0.63f, 0.56f, 0.37f), 0.4f * 128);
Material aluminumMaterial(Color(0.79f, 0.79f, 0.79f), Color(0.7f, 0.7f, 0.7f), Color(0.7f, 0.7f, 0.7f), 0.6f * 128);
Material blueMaterial(Color(0.19f, 0.19f, 0.8f), Color(0.5f, 0.5f, 0.5f), Color(0.5f, 0.5f, 0.5f), 0.4f * 128);
Material brightMaterial(Color(1.0f, 1.0f, 1.0f), Color(1.0f, 1.0f, 1.0f), Color(1.0f, 1.0f, 1.0f), 0.5f * 128);

void transAndRotateZinDir(const BVector3<float> &pos, const BVector3<float> &dir)
{
	glTranslatef(pos.x, pos.y, pos.z);

	float roty = (float)atan2(dir.x, dir.z) * 180.0f / (float)PI;
	float s = sqrtf(dir.x * dir.x + dir.z * dir.z);
	float rotx = (float)atan2(dir.y, s) * 180.0f / (float)PI;

	glRotatef(roty, 0, 1, 0);
	glRotatef(rotx, -1, 0, 0);
}

void drawLine(BVector3<double> u, BVector3<double> v)
{
	glBegin(GL_LINES);
		glVertex3d(u.x, u.y, u.z);
		glVertex3d(v.x, v.y, v.z);
	glEnd();
}


HandManager::HandManager(float x, float y, float z, float pitch, float yaw, float roll, float scale) :
	m_displayStrain(false), m_displayPPS(false)
{

    m_pos[0] = x;
    m_pos[1] = y;
    m_pos[2] = z;

    m_rot[0] = pitch;
    m_rot[1] = yaw;
    m_rot[2] = roll;

    m_scale = scale;

    //!construct vrml objects to be rendered

    Base        = new HandBase(0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, false);

    F1_Knuckle  = new HandKnuckle(0.0f/10.0f, 0.315f/10.0f, -0.005f/10.0f, 0.0f, 0.0f, 0.0f, 1.0f, false);
    F1_Middle   = new HandMiddle(0.0f/10.0f, 0.405f/10.0f, (-0.515f + 0.345f)/10.0f, 180.0f, 0.0f, 0.0f, 1.0f, true);
    F1_Finger   = new HandFinger(0.0f/10.0f, 0.413f/10.0f, -0.96f/10.0f, 222.0f, 0.0f, 0.0f, 1.0f, true);

    F2_Joint    = new HandJoint(0.25f/10.0f, 0.19f/10.0f, 0.0f/10.0f, 0.0f, 0.0f, 0.0f, 1.0f, true);

	F2_Knuckle  = new HandKnuckle(0.25f/10.0f, 0.315f/10.0f, -0.005f/10.0f, 0.0f, 180.0f, 0.0f, 1.0f, true);
    F2_Middle   = new HandMiddle(0.25f/10.0f, 0.405f/10.0f, (0.515f + 0.345f)/10.0f, 0.0f, 0.0f, 180.0f, 1.0f, true);
    F2_Finger   = new HandFingerSpreadable(0.25f/10.0f, 0.413f/10.0f, 1.47f/10.0f, 42.0f, 0.0f, 180.0f, 1.0f, true);

    F3_Joint    = new HandJoint(-0.25f/10.0f, 0.19f/10.0f, 0.0f/10.0f, 0.0f, 0.0f, 0.0f, 1.0f, true);

	F3_Knuckle  = new HandKnuckle(-0.25f/10.0f, 0.315f/10.0f, -0.005f/10.0f, 0.0f, 180.0f, 0.0f, 1.0f, true);
    F3_Middle   = new HandMiddle(-0.25f/10.0f, 0.405f/10.0f, (0.515f + 0.345f)/10.0f, 0.0f, 0.0f, 180.0f, 1.0f, true);
    F3_Finger   = new HandFingerSpreadable(-0.25f/10.0f, 0.413f/10.0f, 1.47f/10.0f, 42.0f, 0.0f, 180.0f, 1.0f, true);

    //set offsets for HandFingerSpreadables
    F2_Finger->xOffset = -0.25;
    F3_Finger->xOffset = 0.25;

    //!construct bounding spheres

    F1_ftip      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.0125/1.0f, 0.020/1.0f, F1_Finger);
    F2_ftip      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.0125/1.0f, 0.020/1.0f, F2_Finger);
    F3_ftip      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.0125/1.0f, 0.020/1.0f, F3_Finger);

    F1_fmid      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.001/1.0f, 0.000/1.0f, F1_Finger);
    F2_fmid      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.001/1.0f, 0.000/1.0f, F2_Finger);
    F3_fmid      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.001/1.0f, 0.000/1.0f, F3_Finger);

    F1_fend      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.013/1.0f, -0.0175/1.0f, F1_Finger);
    F2_fend      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.013/1.0f, -0.0175/1.0f, F2_Finger);
    F3_fend      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.013/1.0f, -0.0175/1.0f, F3_Finger);

    F1_mmid      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.005/1.0f, 0.012/1.0f, F1_Middle);
    F2_mmid      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.005/1.0f, 0.012/1.0f, F2_Middle);
    F3_mmid      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.005/1.0f, 0.012/1.0f, F3_Middle);

    F1_mend      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.005/1.0f, -0.005/1.0f, F1_Middle);
    F2_mend      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.005/1.0f, -0.005/1.0f, F2_Middle);
    F3_mend      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, 0.005/1.0f, -0.005/1.0f, F3_Middle);

    F1_mfar      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.005/1.0f, -0.04/1.0f, F1_Middle);
    F2_mfar      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.005/1.0f, -0.04/1.0f, F2_Middle);
    F3_mfar      = new BoundingSphere(0.012f/1.0f, 0.0/1.0f, -0.005/1.0f, -0.04/1.0f, F3_Middle);
/*
    F1_ftip      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.0125/10.0f, 0.020/10.0f, F1_Finger);
    F2_ftip      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.0125/10.0f, 0.020/10.0f, F2_Finger);
    F3_ftip      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.0125/10.0f, 0.020/10.0f, F3_Finger);

    F1_fmid      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.001/10.0f, 0.000/10.0f, F1_Finger);
    F2_fmid      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.001/10.0f, 0.000/10.0f, F2_Finger);
    F3_fmid      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.001/10.0f, 0.000/10.0f, F3_Finger);

    F1_fend      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.013/10.0f, -0.0175/10.0f, F1_Finger);
    F2_fend      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.013/10.0f, -0.0175/10.0f, F2_Finger);
    F3_fend      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.013/10.0f, -0.0175/10.0f, F3_Finger);

    F1_mmid      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.005/10.0f, 0.012/10.0f, F1_Middle);
    F2_mmid      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.005/10.0f, 0.012/10.0f, F2_Middle);
    F3_mmid      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.005/10.0f, 0.012/10.0f, F3_Middle);

    F1_mend      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.005/10.0f, -0.005/10.0f, F1_Middle);
    F2_mend      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.005/10.0f, -0.005/10.0f, F2_Middle);
    F3_mend      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, 0.005/10.0f, -0.005/10.0f, F3_Middle);

    F1_mfar      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.005/10.0f, -0.04/10.0f, F1_Middle);
    F2_mfar      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.005/10.0f, -0.04/10.0f, F2_Middle);
    F3_mfar      = new BoundingSphere(0.012f/10.0f, 0.0/10.0f, -0.005/10.0f, -0.04/10.0f, F3_Middle);

*/
    //! add bounding spheres to corresponding models

    F1_Finger->addBoundingSphere(F1_ftip);
    F2_Finger->addBoundingSphere(F2_ftip);
    F3_Finger->addBoundingSphere(F3_ftip);

    F1_Finger->addBoundingSphere(F1_fmid);
    F2_Finger->addBoundingSphere(F2_fmid);
    F3_Finger->addBoundingSphere(F3_fmid);

    F1_Finger->addBoundingSphere(F1_fend);
    F2_Finger->addBoundingSphere(F2_fend);
    F3_Finger->addBoundingSphere(F3_fend);

    F1_Middle->addBoundingSphere(F1_mmid);
    F2_Middle->addBoundingSphere(F2_mmid);
    F3_Middle->addBoundingSphere(F3_mmid);

    F1_Middle->addBoundingSphere(F1_mend);
    F2_Middle->addBoundingSphere(F2_mend);
    F3_Middle->addBoundingSphere(F3_mend);

    F1_Middle->addBoundingSphere(F1_mfar);
    F2_Middle->addBoundingSphere(F2_mfar);
    F3_Middle->addBoundingSphere(F3_mfar);

    //! add references to bounding spheres in vector m_b_spheres

    m_b_spheres.push_back(F1_ftip);
    m_b_spheres.push_back(F2_ftip);
    m_b_spheres.push_back(F3_ftip);

    m_b_spheres.push_back(F1_fmid);
    m_b_spheres.push_back(F2_fmid);
    m_b_spheres.push_back(F3_fmid);

    m_b_spheres.push_back(F1_fend);
    m_b_spheres.push_back(F2_fend);
    m_b_spheres.push_back(F3_fend);

    m_b_spheres.push_back(F1_mmid);
    m_b_spheres.push_back(F2_mmid);
    m_b_spheres.push_back(F3_mmid);

    m_b_spheres.push_back(F1_mend);
    m_b_spheres.push_back(F2_mend);
    m_b_spheres.push_back(F3_mend);

    m_b_spheres.push_back(F1_mfar);
    m_b_spheres.push_back(F2_mfar);
    m_b_spheres.push_back(F3_mfar);

    //! add references to models in vector m_parts

    m_parts.push_back(Base);

    m_parts.push_back(F1_Knuckle);
    m_parts.push_back(F1_Middle);
    m_parts.push_back(F1_Finger);

    m_parts.push_back(F2_Joint);
    m_parts.push_back(F2_Knuckle);
    m_parts.push_back(F2_Middle);
    m_parts.push_back(F2_Finger);

    m_parts.push_back(F3_Joint);
    m_parts.push_back(F3_Knuckle);
    m_parts.push_back(F3_Middle);
    m_parts.push_back(F3_Finger);


	// Kinematics testing
	//m_kinematics.test();

	m_showKinematics = false;

	createSmallSphere();
	smooth = gluNewQuadric();
	flat = gluNewQuadric();

	for (int i = 0; i < NUM_FINGER_TIPS; i++)
		fingerTips[i] = VisualFingerTip::getInstance(i);

	palm = VisualPalm::getInstance();
}

//! memory management - clean allocated memory
HandManager::~HandManager() {
	delete Base;
    delete F1_Knuckle;
    delete F1_Middle;
    delete F1_Finger;
    delete F2_Joint;
    delete F2_Knuckle;
    delete F2_Middle;
    delete F2_Finger;
    delete F3_Joint;
    delete F3_Knuckle;
    delete F3_Middle;
    delete F3_Finger;

    delete F1_ftip;
    delete F2_ftip;
    delete F3_ftip;

    delete F1_fmid;
    delete F2_fmid;
    delete F3_fmid;

    delete F1_fend;
    delete F2_fend;
    delete F3_fend;

    delete F1_mmid;
    delete F2_mmid;
    delete F3_mmid;

    delete F1_mend;
    delete F2_mend;
    delete F3_mend;

    delete F1_mfar;
    delete F2_mfar;
    delete F3_mfar;

//    if(m_pos)
//        delete[] m_pos;
//    if(m_rot)
//        delete[] m_rot;
//    if(rot_prev)
//        delete[] rot_prev;
//    if(transforms)
//        delete[] transforms;
//    if(transmats)
//        delete[] transmats;

    m_parts.clear();
    m_b_spheres.clear();
    selected_parts.clear();
}

//! store the state of rotations of models whenever no collision is detected
void HandManager::Cache() {

    float* temp_rot;
    int counter = 0;

    //! store rotations pertaining to parts
    for(unsigned int i = 0; i < m_parts.size(); i++) {

        temp_rot = (m_parts[i])->getRotation();
        rot_prev[counter++] = temp_rot[0];
        rot_prev[counter++] = temp_rot[1];
        rot_prev[counter++] = temp_rot[2];
    }

    rot_prev[counter++] = F1_Finger->rotVal;
    rot_prev[counter++] = F2_Finger->rotVal;
    rot_prev[counter++] = F3_Finger->rotVal;

    counter = 0;
    GLdouble* temppos;
    GLdouble* tempmat;

    //! store transformations pertaining to bounding spheres
    for(unsigned int i = 0; i < m_b_spheres.size(); i++) {
        temppos = m_b_spheres[i]->getPositionTransformed();
        tempmat = m_b_spheres[i]->getTransformation();

        transforms[counter] = temppos;
        transmats[counter++] = tempmat;
    }
}

//! revert back to the last cached state if a collision is detected, restoring previous rotations
void HandManager::Revert() {

    float temp_rot[3];
    int counter = 0;

    ///! restore rotations pertaining to parts
    for(unsigned int i = 0; i < m_parts.size(); i++) {

        temp_rot[0] = rot_prev[counter++];
        temp_rot[1] = rot_prev[counter++];
        temp_rot[2] = rot_prev[counter++];

        (m_parts[i])->setRotation(temp_rot);
    }

    F1_Finger->rotVal = rot_prev[counter++];
    F2_Finger->rotVal = rot_prev[counter++];
    F3_Finger->rotVal = rot_prev[counter++];

    counter = 0;
    GLdouble temp[16];

    //! restore transformations pertaining to bounding spheres
    for(unsigned int i = 0; i < m_b_spheres.size(); i++) {

        for(int j = 0; j < 4; j++)
            temp[j] = (transforms[counter])[j];

        m_b_spheres[i]->setPositionTransformed(temp);

        for(int j = 0; j < 16; j++)
            temp[j] = (transmats[counter])[j];

        m_b_spheres[i]->setTransformation(temp);

        counter++;
    }
}

//! match a color-picked pixel with it's corresponding VRMLModel
VRMLModel* HandManager::FindSelectedPart(const unsigned char* pixel) {
    //go through all parts and find the one that matches
    for(unsigned i = 0; i < m_parts.size(); i++)
    {
       if((m_parts[i])->getR() == pixel[0] &&
           (m_parts[i])->getG() == pixel[1] &&
            (m_parts[i])->getB() == pixel[2] )
       {
            //found a match, return it
            return m_parts[i];
       }
    }

    //if no match is found
    return NULL;
}

//! check to see if a VRMLModel is within the vector of selected parts
bool HandManager::isSelected(const VRMLModel* toCheck) {
    for(unsigned int i = 0; i < selected_parts.size(); i++) {
        if(toCheck == selected_parts[i]) {
            return true;
        }
    }

    return false;
}

//! if one part of the finger is selected, make sure all parts of that finger
//! are considered selected as well
void HandManager::SelectUpdate() {

    if(F1_Finger->isSelected())
        F1_Middle->setSelected(true);

    if(F1_Middle->isSelected())
        F1_Finger->setSelected(true);

    if(F2_Finger->isSelected())
        F2_Middle->setSelected(true);

    if(F2_Middle->isSelected())
        F2_Finger->setSelected(true);

    if(F3_Finger->isSelected())
        F3_Middle->setSelected(true);

    if(F3_Middle->isSelected())
        F3_Finger->setSelected(true);

    //update the vector of currently selected parts
    UpdateSelection();
}

//! if one part of the finger is deselected, make sure all parts of that finger
//! are considered deselected as well
void HandManager::DeselectUpdate() {

    if(F1_Finger->isSelected() && !F1_Middle->isSelected())
        F1_Finger->setSelected(false);
    if(!F1_Finger->isSelected() && F1_Middle->isSelected())
        F1_Middle->setSelected(false);

    if(F2_Finger->isSelected() && !F2_Middle->isSelected())
        F2_Finger->setSelected(false);
    if(!F2_Finger->isSelected() && F2_Middle->isSelected())
        F2_Middle->setSelected(false);

    if(F3_Finger->isSelected() && !F3_Middle->isSelected())
        F3_Finger->setSelected(false);
    if(!F3_Finger->isSelected() && F3_Middle->isSelected())
        F3_Middle->setSelected(false);

    //update the vector of currently selected parts
    UpdateSelection();
}

//! records which models are currently considered to be selected
void HandManager::UpdateSelection() {

    selected_parts.clear();

    if(F1_Finger->isSelected())
        selected_parts.push_back(F1_Finger);

    if(F1_Middle->isSelected())
        selected_parts.push_back(F1_Middle);

    if(F2_Finger->isSelected())
        selected_parts.push_back(F2_Finger);

    if(F2_Middle->isSelected())
        selected_parts.push_back(F2_Middle);

    if(F3_Finger->isSelected())
        selected_parts.push_back(F3_Finger);

    if(F3_Middle->isSelected())
        selected_parts.push_back(F3_Middle);
}

//! moves the selected models in unison
void HandManager::MoveSelected(int xMove, int yMove) {
    //! move each model one at a time the same amount
    for(unsigned int i = 0; i < selected_parts.size(); i++) {
        Move(selected_parts[i], xMove, yMove);
    }
}

//! moves one component and accounts for range-of-motion, joint stops, etc.
//! the Movement is based on delta mouse movement, so if one wanted to extend
//! this function to degree movement there would need to be some experimenting and
//! tweaking of parameters involved.  This function essentially provides rigid-body
//! animation of the hand components.
void HandManager::Move(VRMLModel* selected, int xMove, int yMove) {
    xMove /= 5;
    yMove /= 5;

    float* rot1 = F2_Joint->getRotation();
    float* rot2 = F2_Knuckle->getRotation();
    float* rot3 = F3_Joint->getRotation();
    float* rot4 = F3_Knuckle->getRotation();

    float* middleRot1 = F1_Middle->getRotation();
    float* middleRot2 = F2_Middle->getRotation();
    float* middleRot3 = F3_Middle->getRotation();

    float* fingerRot1 = F1_Finger->getRotation();
    float* fingerRot2 = F2_Finger->getRotation();
    float* fingerRot3 = F3_Finger->getRotation();

    if(selected == F2_Joint || selected == F2_Knuckle) {
        rot1[1] += xMove;
        rot2[1] += xMove;
        rot3[1] -= xMove;
        rot4[1] -= xMove;

        middleRot2[1] -= xMove;
        fingerRot2[1] += xMove;
        middleRot3[1] += xMove;
        fingerRot3[1] -= xMove;
    } else if (selected == F3_Joint || selected == F3_Knuckle) {
        rot1[1] -= xMove;
        rot2[1] -= xMove;
        rot3[1] += xMove;
        rot4[1] += xMove;

        middleRot2[1] += xMove;
        fingerRot2[1] -= xMove;
        middleRot3[1] -= xMove;
        fingerRot3[1] += xMove;
    }

    if(selected == F1_Middle || selected == F1_Finger) {
        middleRot1[0] -= yMove;
        // relative rotation
        F1_Finger->rotVal -= yMove * .95f;
        // local rotation
        fingerRot1[0] -= yMove * 48.0f/140.0f;
    }
    if(selected == F2_Middle || selected == F2_Finger) {
        middleRot2[0] -= yMove;
        // relative rotation
        F2_Finger->rotVal += yMove * .95f;
        // local rotation X axis
        fingerRot2[0] -= yMove * 48.0f/140.0f;
    }
    if(selected == F3_Middle || selected == F3_Finger) {
        middleRot3[0] -= yMove;
        // relative rotation
        F3_Finger->rotVal += yMove * .95f;
        // local rotation X axis
        fingerRot3[0] -= yMove * 48.0f/140.0f;
    }

    if(rot1[1] < 0.0) {
        rot1[1] = 0.0;
        middleRot2[1] = 0.0;
        fingerRot2[1] = 0.0;
    } else if(rot1[1] > 180.0) {
        rot1[1] = 180.0;
        middleRot2[1] = 180.0;
        fingerRot2[1] = 180.0;
    }

    if(rot2[1] < 180.0) {
        rot2[1] = 180.0;
    } else if(rot2[1] > 360.0) {
        rot2[1] = 360.0;
    }

    if(rot3[1] > 0.0) {
        rot3[1] = 0.0;
        middleRot3[1] = 0.0;
        fingerRot3[1] = 0.0;
    } else if(rot3[1] < -180.0) {
        rot3[1] = -180.0;
        middleRot3[1] = -180.0;
        fingerRot3[1] = -180.0;
    }

    if(rot4[1] > 180.0) {
        rot4[1] = 180.0;
    } else if(rot4[1] < 0.0) {
        rot4[1] = 0.0;
    }

    if(middleRot1[0] < 180.0) {
        middleRot1[0] = 180.0;
    } else if (middleRot1[0] > 320.0) {
        middleRot1[0] = 320.0;
    }

    if(F1_Finger->rotVal < 0.0) {
        F1_Finger->rotVal = 0.0;
    } else if(F1_Finger->rotVal > 133.0) {
        F1_Finger->rotVal = 133.0;
    }

    if(fingerRot1[0] > 270.0) {
        fingerRot1[0] = 270.0;
    } else if(fingerRot1[0] < 222.0) {
        fingerRot1[0] = 222.0;
    }

    if(middleRot2[0] > 140.0) {
        middleRot2[0] = 140.0;
    } else if (middleRot2[0] < 0.0) {
        middleRot2[0] = 0.0;
    }

    if(F2_Finger->rotVal > 0.0) {
        F2_Finger->rotVal = 0.0;
    } else if(F2_Finger->rotVal < -133.0) {
        F2_Finger->rotVal = -133.0;
    }

    if(fingerRot2[0] > 90.0) {
        fingerRot2[0] = 90.0;
    } else if(fingerRot2[0] < 42.0) {
        fingerRot2[0] = 42.0;
    }

    if(middleRot3[0] > 140.0) {
        middleRot3[0] = 140.0;
    } else if (middleRot3[0] < 0.0) {
        middleRot3[0] = 0.0;
    }

    if(F3_Finger->rotVal > 0.0) {
        F3_Finger->rotVal = 0.0;
    } else if(F3_Finger->rotVal < -133.0) {
        F3_Finger->rotVal = -133.0;
    }

    if(fingerRot3[0] > 90.0) {
        fingerRot3[0] = 90.0;
    } else if(fingerRot3[0] < 42.0) {
        fingerRot3[0] = 42.0;
    }

}

void HandManager::RefreshLocation(float f1, float f2, float f3, float spread) {

    float* rot1 = F2_Joint->getRotation();
    float* rot2 = F2_Knuckle->getRotation();
    float* rot3 = F3_Joint->getRotation();
    float* rot4 = F3_Knuckle->getRotation();

    float* middleRot1 = F1_Middle->getRotation();
    float* middleRot2 = F2_Middle->getRotation();
    float* middleRot3 = F3_Middle->getRotation();

    float* fingerRot1 = F1_Finger->getRotation();
    float* fingerRot2 = F2_Finger->getRotation();
    float* fingerRot3 = F3_Finger->getRotation();


    middleRot1[0] = 180 + f3;
    fingerRot1[0] = ((f3 / 140.0f) * 48.0f) + 222.0f;
    F1_Finger->rotVal = ((f3 / 140.0f) * 133.0f);

    middleRot2[0] = f2;
    fingerRot2[0] = ((f2 / 140.0f) * 48.0f) + 42.0f;
    F2_Finger->rotVal = ((f2 / 140.0f) * -133.0f);

    middleRot3[0] = f1;
    fingerRot3[0] = ((f1 / 140.0f) * 48.0f) + 42.0f;
    F3_Finger->rotVal = ((f1 / 140.0f) * -133.0f);

    rot1[1] = spread;
    rot2[1] = spread + 180;
    middleRot2[1] = -spread;
    fingerRot2[1] = spread;

    rot3[1] = -spread;
    rot4[1] = -spread + 180;
    middleRot3[1] = spread;
    fingerRot3[1] = -spread;
}

float HandManager::getFingerRot1() {
    return (F3_Middle->getRotation())[0];
}

float HandManager::getFingerRot2() {
    return (F2_Middle->getRotation())[0];
}

float HandManager::getFingerRot3() {
    return (F1_Middle->getRotation())[0] - 180.0f;
}

float HandManager::getSpreadRot() {
    return (F2_Joint->getRotation())[1];
}

void HandManager::createSmallSphere()
{
	smallSphere = glGenLists(1);

	glNewList(smallSphere, GL_COMPILE);

	glColor3f(0.1f, 0.1f, 0.1f);

	glutSolidSphere(HAND_KINEMATICS_RADIUS, 32, 32);

	glEndList();
}

// display the hand and all its components
void HandManager::DrawHand(const bool ColorPicking)
{
    glPushMatrix();

    // perform global transformation for all hand components
    glTranslatef(m_pos[0], m_pos[1], m_pos[2]);
    glRotatef(m_rot[0], 1.0, 0.0, 0.0);
    glRotatef(m_rot[1], 0.0, 1.0, 0.0);
    glRotatef(m_rot[2], 0.0, 0.0, 1.0);
    //glScalef(m_scale, m_scale, m_scale);

    // go through and render each part
    for (unsigned i = 0; i < m_parts.size(); i++)
	{

        // set opengl's render color to be the pickable object's color
        if (ColorPicking && m_parts[i]->isPickable())
            glColor3ub(m_parts[i]->getR(), m_parts[i]->getG(), m_parts[i]->getB());

        // render only pickable objects in color-picking mode
        // otherwise, render everything
        if ((ColorPicking && m_parts[i]->isPickable()) || !ColorPicking)
            m_parts[i]->Render();
    }


	if (!ColorPicking)
	{
		// Prepare coordinate system used in the User Manual
	    glPushMatrix();

			// perform global rotation for all hand kinematics (Z axis should be up, not facing forward)
			glRotatef(-90, 1.0, 0.0, 0.0); // replaces //BVector3<double> toOpenGLCoord(const BVector3<double> v) { return BVector3<double>(v.x * 0.01, (v.z - 84/2) * 0.01, -v.y * 0.01 ); }
			glTranslatef(0, 0, -0.0840f/2);

			glLineWidth(6);

			// Prepare encoder positions to use for calculating forward kinematics
			double motorEncPos[4];
			motorEncPos[0] = DEG2RAD(getFingerRot1());
			motorEncPos[1] = DEG2RAD(getFingerRot2());
			motorEncPos[2] = DEG2RAD(getFingerRot3());
			motorEncPos[3] = DEG2RAD(getSpreadRot());
			//printf("enc positions (radians) = %8.3f %8.3f %8.3f %8.3f\n", motorEncPos[0], motorEncPos[1], motorEncPos[2], motorEncPos[3]);

			double innerEncPos[4];
			innerEncPos[0] = motorEncPos[0] / 3;
			innerEncPos[1] = motorEncPos[1] / 3;
			innerEncPos[2] = motorEncPos[2] / 3;

			// Initialize configuration variables for kinematics
			m_kinematics.fillCapTheta(motorEncPos, innerEncPos);


			/* BH8-262 (does not include breakaway)
			double encPos[4];
			encPos[0] = getFingerRot1() * 17500.0f / 140.0f;
			encPos[1] = getFingerRot2() * 17500.0f / 140.0f;
			encPos[2] = getFingerRot3() * 17500.0f / 140.0f;
			encPos[3] = getSpreadRot() * 3150.0f / 180.0f;
			//printf("enc positions = %8.3f %8.3f %8.3f %8.3f\n", encPos[0], encPos[1], encPos[2], encPos[3]);

			// Initialize configuration variables for kinematics
			m_kinematics.fillCapThetaBefore(encPos);*/

			setMaterial(&aluminumMaterial);

			// Draw Zero Position (for picture in manual)
			//drawLine(BVector3<double>(0, 0, 0), BVector3<double>(0.0110, 0, 0));
			//drawLine(BVector3<double>(0, 0, 0), BVector3<double>(0, 0.0110, 0));
			//drawLine(BVector3<double>(0, 0, 0), BVector3<double>(0, 0, 0.0130));

			// Allocate space for transformation matrices
			BMatrix<double> matrices[4];
			matrices[0].setSize(4, 4);
			matrices[1].setSize(4, 4);
			matrices[2].setSize(4, 4);
			matrices[3].setSize(4, 4);

			for (int i = 1; i <= 3; i++)
			{

				// Compute transformation matrices
				BHandKinematics::setD3(0); // D3 adjusted for OpenGL model
				m_kinematics.getBHandMatrices(i, &matrices[0], &matrices[1], &matrices[2], &matrices[3]);

				BVector3<double> endpoints[4];

				if (m_showKinematics)
				{
					// Show spheres for joints and for the end of the finger tip, then connect the joints with cylinders
					setMaterial(&aluminumMaterial);

					for (int m = 0; m < 4; m++)
					{
						// Place endpoints into a vector
						endpoints[m] = BVector3<double>(matrices[m](0, 3), matrices[m](1, 3), matrices[m](2, 3));

						// Draw spheres at each finger joint and for the end of the finger tip
						if (m > 0)
						{
							//setMaterial(&jointMaterial);
							glPushMatrix();
								glTranslatef((float)endpoints[m].x, (float)endpoints[m].y, (float)endpoints[m].z);
								glCallList(smallSphere);
							glPopMatrix();
						}

						// Draw cylinders connecting finger joints
						if (m > 1)
						{
							//setMaterial(&jointMaterial);
							BVector3<double> dir = endpoints[m] - endpoints[m-1];
							BVector3<float> start = endpoints[m-1].toFloat();
							glPushMatrix();
								transAndRotateZinDir(start, dir.toFloat());
								gluCylinder(smooth, HAND_KINEMATICS_RADIUS, HAND_KINEMATICS_RADIUS, dir.Length(), 32, 1);
							glPopMatrix();
						}
					}
				}

				if (m_displayStrain)
				{
					// Display strain at finger tip
					VisualData *vd = VisualData::getInstance();
					{
						setMaterial(&blueMaterial);
						glColor3f(1.0f, 1.0f, 1.0f);
						float strain = -vd->getStrain(i - 1);

						BMatrix<double> temp(4, 1);
						BMatrix<double> po(4, 1); po(0, 0) = 0.012; po(1, 0) = 0; po(2, 0) = 0; po(3, 0) = 1;
						BMatrix<double> px(4, 1); px(0, 0) = 0.012; px(1, 0) = 0; px(2, 0) = 0.12 * strain / 4095.0; px(3, 0) = 1;

						//glEnable(GL_COLOR_MATERIAL);
						glShadeModel(GL_FLAT);
						temp = matrices[3].prod(po); BVector3<double> p0(temp(0, 0), temp(1, 0), temp(2, 0));
						temp = matrices[3].prod(px); BVector3<double> p1(temp(0, 0), temp(1, 0), temp(2, 0)); drawLine(p0, p1);
						//glDisable(GL_COLOR_MATERIAL);
						glShadeModel(GL_SMOOTH);
					}
				}

				// Display finger tips
				if (m_displayPPS)
				{
					setMaterial(&brightMaterial);
					fingerTips[i - 1]->render(matrices[3]);

					// Display palm
					if (i == 3)
						palm->render(matrices[0]);
				}


				// Compute transformation matrices a second time to adjust finger endpoint depth offset
				/*BHandKinematics::setD3(0.0072); // D3 adjusted for OpenGL model
				m_kinematics.getBHandMatrices(i, &matrices[0], &matrices[1], &matrices[2], &matrices[3]);

				for (int m = 0; m < 4; m++)
				{
					// Optional, draw coordinate frames
					//if (m == 3) // Show tool frame only
					{
						setMaterial(&blueMaterial);
						glColor3f(0.7f, 0.7f, 1.0f);
						BMatrix<double> temp(4, 1);
						BMatrix<double> po(4, 1); po(0, 0) =  0; po(1, 0) =  0; po(2, 0) =  0; po(3, 0) = 1;
						BMatrix<double> px(4, 1); px(0, 0) = 0.04; px(1, 0) =  0; px(2, 0) =  0; px(3, 0) = 1;
						BMatrix<double> py(4, 1); py(0, 0) =  0; py(1, 0) = 0.04; py(2, 0) =  0; py(3, 0) = 1;
						BMatrix<double> pz(4, 1); pz(0, 0) =  0; pz(1, 0) =  0; pz(2, 0) = 0.04; pz(3, 0) = 1;

						temp = matrices[m].prod(po); BVector3<double> p0(temp(0, 0), temp(1, 0), temp(2, 0));
						temp = matrices[m].prod(px); BVector3<double> p1(temp(0, 0), temp(1, 0), temp(2, 0)); drawLine(p0, p1);
						temp = matrices[m].prod(py); BVector3<double> p2(temp(0, 0), temp(1, 0), temp(2, 0)); drawLine(p0, p2);
						temp = matrices[m].prod(pz); BVector3<double> p3(temp(0, 0), temp(1, 0), temp(2, 0)); drawLine(p0, p3);
					}
				}*/
			}
		glPopMatrix();
	}

	// render bounding spheres if enabled (#define COLLISION_DEBUG) at the top of this file
	if (COLLISION_DEBUG)
	{
		for (unsigned i = 0; i < m_b_spheres.size(); i++)
			m_b_spheres[i]->Render();
	}

	// perform collision detection for all the bounding spheres and resolve accordingly
	for (unsigned int i = 0; i < m_b_spheres.size(); i++)
	{
		for (unsigned int j = i + 1; j < m_b_spheres.size(); j++)
		{
			if (!m_b_spheres[i]->hasSameParent(m_b_spheres[j]) && m_b_spheres[i]->checkIntersection(m_b_spheres[j]))
			{
				Revert();
				glPopMatrix();
				return;
			}
		}
	}

	setMaterial(&aluminumMaterial);
	glLineWidth(1);

	// no collision found, save the last successfully rendered state
	Cache();
	glPopMatrix();
}

