#include "BoundingSphere.h"
#include <stdio.h>
#include <cmath>

BoundingSphere::BoundingSphere(float radius, GLdouble x, GLdouble y, GLdouble z, VRMLModel* parent) {

    m_position[0] = x;
    m_position[1] = y;
    m_position[2] = z;
    m_position[3] = 1.0;

    m_radius = radius;
    m_parent = parent;
}

BoundingSphere::~BoundingSphere() {
}

//! get functions

float BoundingSphere::getRadius() {
    return m_radius;
}

GLdouble* BoundingSphere::getPositionTransformed() {
    return m_position_transformed;
}

GLdouble* BoundingSphere::getTransformation() {
    return m_transform;
}

VRMLModel* BoundingSphere::getParent() {
    return m_parent;
}

//! set functions

void BoundingSphere::setPositionTransformed(const GLdouble* pos) {
    m_position_transformed[0] = pos[0];
    m_position_transformed[1] = pos[1];
    m_position_transformed[2] = pos[2];
    m_position_transformed[3] = pos[3];
}

void BoundingSphere::setTransformation(const GLdouble* trans) {
    for(int i = 0 ; i < 16; i++) {
        m_transform[i] = trans[i];
    }
}

//! checks to see if two bounding spheres have the same parent and returns
bool BoundingSphere::hasSameParent(BoundingSphere* toCheck) {
    return toCheck->getParent() == m_parent;
}

//! renders a bounding sphere with the correct transform applied to it
void BoundingSphere::Render() const {

    glPushMatrix();
        glLoadMatrixd(m_transform);
        glTranslated(m_position[0], m_position[1], m_position[2]);
        glutWireSphere(m_radius, 10, 10);
    glPopMatrix();
}

//! gets the current transformation of the parent VRMLModel and updates the BoundingSphere's parameters
void BoundingSphere::updateTransformation(const GLdouble* transformation) {

    //deep copy for the transformation matrix
    for(int i = 0; i < 16; i++) {
        m_transform[i] = transformation[i];
    }

    //openGL is column-major order
    //recalculate the transformed position of the bounding sphere
    m_position_transformed[0] = transformation[0] * m_position[0] + transformation[4] * m_position[1] + transformation[8] * m_position[2] + transformation[12] * m_position[3];
    m_position_transformed[1] = transformation[1] * m_position[0] + transformation[5] * m_position[1] + transformation[9] * m_position[2] + transformation[13] * m_position[3];
    m_position_transformed[2] = transformation[2] * m_position[0] + transformation[6] * m_position[1] + transformation[10] * m_position[2] + transformation[14] * m_position[3];
    m_position_transformed[3] = transformation[3] * m_position[0] + transformation[7] * m_position[1] + transformation[11] * m_position[2] + transformation[15] * m_position[3];
}

//! checks to see if two bounding spheres intersected with one another
bool BoundingSphere::checkIntersection(BoundingSphere* toCheck) {

    //get transformed position
    GLdouble* toCheck_pos = toCheck->getPositionTransformed();
    //combine radii
    double radius = toCheck->getRadius() + m_radius;
    //get distance of transformed positions
    double distance = sqrt(pow((toCheck_pos[0] - m_position_transformed[0]), 2) + pow((toCheck_pos[1] - m_position_transformed[1]), 2) + pow((toCheck_pos[2] - m_position_transformed[2]), 2));

    //scale the bounding sphere's radius slightly as there (basically tweaking glut's glutWireSphere settings)
    //return (distance < radius * 10.0); // TODO: earlier version
    return (distance < radius * 1.01);
}
