#include "VisualControl.h"

#include "VisualPanel.h"
#include "VisualData.h"
#include "VisualFingerTip.h"
#include "VisualPalm.h"

#include "BHand.h"


extern bool use280Config;


VisualControl::VisualControl(VisualPanel *visualPanel, BHand *bhand) :
	m_visualPanel(visualPanel), m_hand(bhand),
	m_realtimeMode(false), m_realtimeErrorFlag(false),
	m_newSpreadPositionFlag(false), m_newSpreadPosition(0),
	m_updateStrain(false), m_updatePPS(false),
	m_visualMode(BHVisualVelocityMode),
	m_desVisualMode(BHVisualVelocityMode)
{
	m_visualData = VisualData::getInstance(m_hand);

	for (int i = 0; i < NUM_FINGER_TIPS; i++)
		m_fingerTips[i] = new VisualFingerTip(i);

	m_palm = new VisualPalm();

#ifdef CHECK_SERVO_RATE
	m_servoLoopCount = 0;
	servoLoopTmr.Start();
	m_elapsed = 0;
#endif
}

VisualControl::~VisualControl()
{
	StopRealTimeMode();

	delete m_palm;

	for (int i = 0; i < NUM_FINGER_TIPS; i++)
		delete m_fingerTips[i];

	VisualData::freeInstance();
}


void VisualControl::setParameters(double velPGain, double velDGain, long realtimeGain, long LCVMultiplier)
{
	this->velPGain = velPGain;
	this->velDGain = velDGain;
	this->realtimeGain = realtimeGain;
	this->LCVMultiplier = LCVMultiplier;
}

void VisualControl::setVisualMode(BHVisualMode visualMode)
{
	m_desVisualMode = visualMode;
}

void VisualControl::setOpenGLPos(float *pos)
{
	// Save desired OpenGL positions
	m_fastMutex.lock();
	for (unsigned int i = 0; i < 4; i++)
		m_opengl_pos[i] = pos[i];
	m_fastMutex.unlock();
	//printf("setOpenGLPos %3.3f %3.3f %3.3f %3.3f\n", opengl_pos[0], opengl_pos[1], opengl_pos[2], opengl_pos[3]);
}


void VisualControl::StartRealTimeMode()
{
	// Check to make sure that RealTime mode isn't already started
	if (m_realtimeMode)
		return;

	while (m_thread.isRunning()); // Wait for thread to complete execution
	while (m_realtimeErrorFlag); // Poll error flag to be sure any errors have been processed

	// Start RealTime visual control thread
	m_thread.setName("BHand Visual Control");
	m_thread.setPriority(Poco::Thread::PRIO_HIGHEST);

	// Enable thread to run RealTime mode
	m_realtimeMode = true;

	m_thread.start(*this);
}

void VisualControl::StopRealTimeMode()
{
	if (m_realtimeMode)
	{
		// Request exit from servoloop
		m_realtimeMode = false;
	}
	// Wait for thread to complete execution
	while (m_thread.isRunning()); // maybe better for Ubuntu (tryJoin(0) has blocked execution before in Ubuntu 9.10)
}

bool VisualControl::PrepareRT()
{
	// Reset mode
	m_hand->pCallback = NULL;
	m_hand->syncMode = BHMODE_SYNC;
	m_hand->RTAbort();

	// Extra code for testing breakaway
	//for (int i = 0; i < 4; i++)
	//{
	//	char m[2] = "1";
	//	m[0] += i;
	//	int bdat;

	//	m_hand->Set(m, "BDAT", 6500); // experimenting for finger 1
	//	//m_hand->Set(m, "BDAT", 6500); // works well for finger 2
	//	//m_hand->Set(m, "BDAT", 7000); // works well for finger 3
	//	m_hand->Get(m, "BDAT", &bdat);
	//	wxPuts(wxString::Format(wxT("BDAT[%d] = %d"), i, bdat));
	//}

	// Setup for RealTime mode - returns true iff setup successful

	bool LCV = true;          // loop control velocity
	int LCVC = LCVMultiplier; // loop control velocity coefficient
	bool LCPG = true;         // loop control proportional gain
	bool LCT = false;         // loop control torque

	bool LFV = true;          // loop feedback velocity
	int  LFVC = 1;            // loop feedback velocity coefficient
	bool LFS = false;         // loop feedback strain
	bool LFAP = true;         // loop feedback absolute position
	bool LFDP = false;        // loop feedback delta position
	int  LFDPC = 1;           // loop feedback delta position coefficient

	bool LFBP = false;        // loop feedback breakaway position not needed
	bool LFAIN = false;       // loop feedback analog inputs not needed
	bool LFDPD = false;       // loop feedback delta position discard not needed
	bool LFT = false;         // loop feedback temperature not needed


	// Set RealTime mode flags
	if ((m_hand->RTSetFlags("123S",
		LCV, LCVC, LCPG, LCT,
		LFV, LFVC, LFS, LFAP, LFDP, LFDPC,
		LFBP, LFAIN, LFDPD, LFT)))
		return false;

	if (use280Config)
	{
		// Set Motor mode
		switch (m_visualMode)
		{
			case BHVisualVelocityMode: { m_hand->Set("123S", "LCV",  true); m_hand->Set("123S", "LCP", false); m_hand->Set("123S", "LCT", false); break; }
			case BHVisualPositionMode: { m_hand->Set("123S", "LCV", false); m_hand->Set("123S", "LCP",  true); m_hand->Set("123S", "LCT", false); break; }
			case BHVisualTorqueMode:   { m_hand->Set("123S", "LCV", false); m_hand->Set("123S", "LCP", false); m_hand->Set("123S", "LCT",  true); break; }
		}

		for (int i = 0; i < 4; i++)
		{
			// Get present encoder positions
			char m[] = "1"; m[0] += i;
			int actualPosition;
			if (m_hand->Get(m, "P", &actualPosition))
				return false;

			// Save present encoder positions for fingers and spread
			if (i < 3)
				m_refPos[i] = fingerEncoderTicksToRadians(actualPosition);
			else
				m_refPos[i] = spreadEncoderTicksToRadians(actualPosition);

			//printf("actualPositions[%d] = %d, m_refPos = %3.3f, opengl_pos = %3.3f\n", i, actualPosition, m_refPos[i], opengl_pos[i]);

			// Warning - this is extremely dangerous since this disables an important safety feature that keeps motors from overheating and burning up
			//m_hand->Set(m, "TSTOP", 0); // don't let the motors leave control mode (position, velocity, or torque mode)

			// Set these variables for implementing velocity mode with torque control (not anymore)
			//torPIDValue[i] = 0;
			//torLastPIDValue[i] = 0;
			//fingerLastVelError[i] = 0;
		}

		m_hand->Set("GS", "LFPPS", m_updatePPS ? 1 : 0);
	}
	else if (m_updateStrain)
	{
		m_hand->Set("G", "LFS", 1);
	}


	if (!use280Config && m_visualMode == BHVisualVelocityMode)
	{
		// Start RealTime mode with all motors except for spread
		if (m_hand->RTStart("123", BHMotorTorqueLimitProtect))
			return false;
	}
	else
	{
		// Position or Torque mode
		// Start RealTime mode for all motors
		if (m_hand->RTStart("123S", BHMotorTorqueLimitProtect))
			return false;
	}

	for (unsigned int i = 0; i < 4; i++)
		m_hand->RTSetGain('1' + i, realtimeGain); // only has an effect for 262 hand

	return true;
}

void VisualControl::run()
{
	bool errorFlag = false; // Used to keep track of errors while running RealTime mode

	// Initialize visual mode variables
	m_visualMode = m_desVisualMode;

	//float prop_gain = (float)velPGain;
	//float deriv_gain = (float)velDGain;

	// Prepare RealTime mode
	if (!PrepareRT())
	{
		errorFlag = true;
		m_realtimeMode = false;
	}

	dtTmr.Start();

	while (m_realtimeMode)
	{
		// Check for new spread position flag (may interrupt RealTime mode and jump to new spread position)
		if (!use280Config && m_newSpreadPositionFlag)
		{
			if (m_visualMode == BHVisualVelocityMode)
			{
				// Execute Move spread position
				// Abort RT mode
				if (m_hand->RTAbort())
				{
					errorFlag = true;
					m_realtimeMode = false;
					break;
				}

				if (m_hand->GoToPosition("4", m_newSpreadPosition))
				{
					errorFlag = true;
					m_realtimeMode = false;
					break;
				}

				// Re-enter RT mode
				if (!PrepareRT())
				{
					errorFlag = true;
					m_realtimeMode = false;
					break;
				}
			}

			// Clear new spread position flag
			m_newSpreadPositionFlag = false;
		}

		// Do main update
		if (!RTPositionUpdate())
		{
			// An error occurred so stop
			errorFlag = true;
			m_realtimeMode = false;
			break;
		}

		// Do update to get sensor data from hand
		if (m_updateStrain && m_visualData->update())
		{
			// An error occurred so stop
			errorFlag = true;
			m_realtimeMode = false;
			break;
		}

		// Update finger tip sensor data from hand
		if (m_updatePPS)
			for (int i = 0; i < NUM_FINGER_TIPS; i++)
				m_fingerTips[i]->update(m_hand);

		// Update palm sensor data from hand
		if (m_updatePPS)
			m_palm->update(m_hand);

#ifdef CHECK_SERVO_RATE
		m_serviceMutex.lock();
		m_elapsed = servoLoopTmr.GetElapsedInSecs();
		m_servoLoopCount++;
		m_serviceMutex.unlock();
#endif
	}
	if (!errorFlag)
	{
		m_hand->RTAbort();
		m_hand->Command("T");
	}
	else
		m_realtimeErrorFlag = errorFlag;
}

bool VisualControl::RTPositionUpdate()
{
	int result;
	float ap[4];

	double opengl_pos[4];

	// Get a copy of desired OpenGL positions
	m_fastMutex.lock();
	for (unsigned int i = 0; i < 4; i++)
		opengl_pos[i] = m_opengl_pos[i];
	m_fastMutex.unlock();

	if (m_visualMode == BHVisualVelocityMode)
	{
		for (unsigned int i = 0; i < (use280Config ? 4 : 3); i++)
		{
			int actualVel = m_hand->RTGetVelocity('1' + i);
			if (i < 3)
				ap[i] = fingerEncoderTicksToRadians(m_hand->RTGetPosition('1' + i));
			else
				ap[i] = spreadEncoderTicksToRadians(m_hand->RTGetPosition('1' + i));

			if (use280Config)
			{
				float velocity;
				velocity = (opengl_pos[i] - ap[i]) * velPGain;

				actualVel = -actualVel; // flip sign on encoders

				// Set desired motor velocities
				if (i < 3)
					m_hand->RTSetVelocity('1' + i, fingerRadiansToEncoderTicks(velocity * 0.05)); // scaling constant for 280
				else
					m_hand->RTSetVelocity('1' + i, spreadRadiansToEncoderTicks(velocity * 0.05)); // scaling constant for 280
			}
			else
			{
				float velocity;
				velocity = (opengl_pos[i] - ap[i]) * velPGain * (-16000);
				velocity -= actualVel * velDGain;

				// Set desired motor velocities
				m_hand->RTSetVelocity('1' + i, -(int)velocity);
				//printf("opengl_pos[%d]");
				//printf("opengl_pos[%d]=%3.3f, ap[%d]=%3.3f, velocity=%3.3f, finalvel=%d\n", i, opengl_pos[i], i, ap[i], velocity, -(int)velocity );
			}

			//wxPuts(wxString::Format(wxT("opengl_pos[%d]=%3.3f, ap[%d]=%3.3f, velocity=%3.3f, finalvel=%d"), i, opengl_pos[i], i, ap[i], velocity, -fingerRadiansToEncoderTicks(velocity) ));
			//printf("opengl_pos[%d]=%3.3f, ap[%d]=%3.3f, velocity=%3.3f, finalvel=%d\n", i, opengl_pos[i], i, ap[i], velocity, -fingerRadiansToEncoderTicks(velocity) );


			//if (i == 0)
			//{
			//	int bp = m_hand->RTGetBreakawayPosition('1' + i);
			//	wxPuts(wxString::Format(wxT("set vel[%d]=%4d, p=%4d,bp=%4d"), i, (int)velocity, (int)ap[i], bp));
			//}

		}

		// Do the control loop update
		if ((result = m_hand->RTUpdate(true, true)))
			return false;
	}
	else if (m_visualMode == BHVisualPositionMode)
	{
		float dt = (float)dtTmr.GetElapsedInSecs();

		// Force min dt to be 0.001 seconds
		if (dt < 0.001f)
			dt = 0.001f;

		for (unsigned int i = 0; i < 4; i++)
		{
			// 1) Movement with constant velocity
			/*float DELTA_POS = 0.001;

			if (m_refPos[i] - DELTA_POS > opengl_pos[i])
				m_refPos[i] -= DELTA_POS;
			else if (m_refPos[i] + DELTA_POS < opengl_pos[i])
				m_refPos[i] += DELTA_POS;*/


			// 2) Exponential filtering
			//m_refPos[i] = m_refPos[i] * 0.95 + opengl_pos[i] * 0.03;

			// 3) Adjust movement amount to be proportional to position error scaled by dt
			m_refPos[i] -= ((float)(m_refPos[i] - opengl_pos[i]) * dt) * 1.7;

			// 4) Combination of 1 and 3
			float MIN_AMOUNT = 0.00125f;

			float amount = ((float)(m_refPos[i] - opengl_pos[i]) * dt) * 1.7;
			if (amount > 0)
			{
				if (amount < MIN_AMOUNT)
					amount = MIN_AMOUNT;
				if (m_refPos[i] - amount > opengl_pos[i])
					m_refPos[i] -= amount;
			}
			else if (amount < 0)
			{
				if (amount > -MIN_AMOUNT)
					amount = -MIN_AMOUNT;
				if (m_refPos[i] - amount < opengl_pos[i])
					m_refPos[i] -= amount;
			}

			// Set desired motor positions
			if (i < 3)
				m_hand->RTSetPosition('1' + i, fingerRadiansToEncoderTicks(m_refPos[i]));
			else
				m_hand->RTSetPosition('1' + i, spreadRadiansToEncoderTicks(m_refPos[i]));

		}

		dtTmr.Start();

		// Do the control loop update
		if ((result = m_hand->RTUpdate(true, false))) // no feedback
			return false;
	}
	else if (m_visualMode == BHVisualTorqueMode)
	{
		// Safety check to ensure minimum servorate
		if (dtTmr.GetElapsedInSecs() < 0.015)
		{
			for (int i = 0; i < 4; i++)
			{
				ap[i] = (i < 3) ?
					fingerEncoderTicksToRadians(m_hand->RTGetPosition('1' + i)) :
					spreadEncoderTicksToRadians(m_hand->RTGetPosition('1' + i));

				// Finger Gains - these gains were really lowered to deal with varying servo rates and friction
				torPGainFinger = 500;
				torDGainFinger = 3.0;
				//torPGainFinger = 2400;
				//torDGainFinger = 4;

				// Spread Gains
				torPGainSpread = 2400/2;
				torDGainSpread = 36/2;


				float torque;
				if (i < 3)
				{
					// Another torque control method (much more difficult to tune)
					/*float actualVel = (float)m_hand->RTGetVelocity('1' + i);

					// Set desired motor velocities
					float velocity;
					ap[i] = fingerEncoderTicksToRadians(m_hand->RTGetPosition('1' + i));
					velocity = (opengl_pos[i] - ap[i]) * prop_gain;
					//float desVel = fingerRadiansToEncoderTicks(velocity * 0.05) * 0.7;
					float desVel = fingerRadiansToEncoderTicks(velocity * 0.05) * 1;

					float fingerVelError = actualVel - desVel;

					//float fingerVelPGain = 7; //0.8
					//float fingerVelDGain = 2;//1.5f; // 2.4

					//float fingerVelPGain = 1.0f;
					//float fingerVelDGain = 1.5f;


					//if (i == 0)
					//	printf("actualVel = %3.3f, desVel = %3.3f, fingerVelError = %3.3f\n", actualVel, desVel, fingerVelError);

					torPIDValue[i] = torLastPIDValue[i] +
						fingerVelPGain * fingerVelError +
						fingerVelDGain * (fingerVelError - fingerLastVelError[i]);

					torLastPIDValue[i] = torPIDValue[i];
					fingerLastVelError[i] = fingerVelError;

					torque = torPIDValue[i];

					// Limit torque (fingers may not be able to get off their stops but user can user position or velocity mode for that)
					if (torque > 1200)
						torque -= 1200;
					else if (torque < -1200)
						torque += 1200;*/

					// Original torque control method
					int actualVel = m_hand->RTGetVelocity('1' + i);
					float posError = ap[i] - opengl_pos[i];
					if (posError > 3.14f/2)
						posError = 3.14f/2;
					else if (posError < -3.14/2)
						posError = -3.14f/2;
					torque = posError * torPGainFinger;
					torque += actualVel * torDGainFinger;
				}
				else
				{
					int actualVel = m_hand->RTGetVelocity('1' + i);
					torque = (ap[i] - opengl_pos[i]) * torPGainSpread;
					torque += actualVel * torDGainSpread;
				}

				m_hand->RTSetTorque('1' + i, -(int)torque);

				//wxPuts(wxString::Format(wxT("opengl_pos[%d]=%3.3f, ap[%d]=%3.3f, velocity=%3.3f, finalvel=%d"), i, opengl_pos[i], i, ap[i], velocity, -fingerRadiansToEncoderTicks(velocity) ));
				//if (i == 3)
				//	printf("%3.3f - (%3.3f) = %3.3f \n", (opengl_pos[i] - ap[i]) * torPGainSpread, actualVel * torDGainSpread, torque);
					//printf("opengl_pos[%d]=%3.3f, ap[%d]=%3.3f, damping = %f, torque=%f\n", i, opengl_pos[i], i, ap[i], actualVel * torDGainSpread, -torque );
			}
		}
		else
		{
			//printf("dt = %1.3f", dtTmr.GetElapsedInSecs());
			// Minimum servorate not met
			for (int i = 0; i < 4; i++)
				m_hand->RTSetTorque('1' + i, 0);
		}

		dtTmr.Start();

		// Do the control loop update
		if ((result = m_hand->RTUpdate(true, true)))
			return false;

	}

	return true;
}

void VisualControl::MoveSpread(int newSpreadPosition)
{
	if (!m_newSpreadPositionFlag)
	{
		m_newSpreadPosition = newSpreadPosition;
		m_newSpreadPositionFlag = true;
	}
}

void VisualControl::Service() // Called periodically through an idle event
{
#ifdef CHECK_SERVO_RATE
	// Check servo rate
	static double m_lastElapsed = 0;
	m_serviceMutex.lock();
	if (m_elapsed > m_lastElapsed + 1)
	{
		printf("Servo rate = %d\n", m_servoLoopCount);

		m_servoLoopCount = 0;

		m_lastElapsed = m_elapsed;
	}
	m_serviceMutex.unlock();
#endif

	// Check for error flag
	if (m_realtimeErrorFlag)
	{
		printf("VisualControl Error Flag was set and is being serviced\n");
		StopRealTimeMode();
		m_visualPanel->StopRTControl();
		m_hand->Command("T"); // Not expected, so kill power to motors

		m_realtimeErrorFlag = false;
	}

}

