#ifndef VISUAL_CONTROL_H
#define VISUAL_CONTROL_H

#include "Common/tmr.h"

// POCO library includes
#include "Poco/Thread.h"
#include "Poco/Runnable.h"
#include "Poco/Mutex.h"
#include "Poco/Semaphore.h"
#include "Poco/Event.h"

using Poco::Thread;
using Poco::Runnable;
using Poco::Event;
using Poco::Semaphore;
using Poco::FastMutex;

//#define CHECK_SERVO_RATE

#define NUM_FINGER_TIPS 3

enum BHVisualMode
{
	BHVisualVelocityMode = 0,
	BHVisualPositionMode,
	BHVisualTorqueMode,
};


class BHand;
class VisualPanel;
class VisualData;
class VisualFingerTip;
class VisualPalm;

class VisualControl : public Runnable
{
public:
	VisualControl(VisualPanel *visualPanel, BHand *bhand);
	virtual ~VisualControl();

	// Start, Stop, Accessor methods for RealTime
	void StartRealTimeMode();
	void StopRealTimeMode();

	bool InRealTimeMode() { return m_realtimeMode; }

	// Init velocity mode parameters
	void setParameters(double velPGain, double velDGain, long realtimeGain, long LCVMultiplier);

	void setVisualMode(BHVisualMode visualMode);

	// Set hand motor reference positions
	void setOpenGLPos(float *pos);

	void updateStrain(bool update) { m_updateStrain = update; }
	void updatePPS(bool update) { m_updatePPS = update; }

	// Events
	void MoveSpread(int newSpreadPosition);

	void Service();


private:

	bool PrepareRT();
	bool RTPositionUpdate();

	///////////////////////////////////////////////////////////////////////////
	// Private method (required to be implemented to be Runnable)
	///////////////////////////////////////////////////////////////////////////

	void run();

	///////////////////////////////////////////////////////////////////////////
	// Private variables
	///////////////////////////////////////////////////////////////////////////

	VisualPanel *m_visualPanel;
	VisualData *m_visualData;
	BHand *m_hand;

	volatile bool m_realtimeMode;
	volatile bool m_realtimeErrorFlag;

	Thread m_thread;
	FastMutex m_fastMutex;

#ifdef CHECK_SERVO_RATE
	volatile int m_servoLoopCount;
	HiResTimer servoLoopTmr;
	volatile double m_elapsed;

	FastMutex m_serviceMutex;
#endif

	// Visual State
	float m_opengl_pos[4];

	// Variables for Event Handlers
	volatile bool m_newSpreadPositionFlag;
	volatile int m_newSpreadPosition;

	// Feedback for optional strain gauges and pressure profile sensors
	volatile bool m_updateStrain;
	volatile bool m_updatePPS;

	// Timer
	HiResTimer dtTmr;

	///////////////////////////////////////////////////////////////////////////
	// Controller State (RealTime Mode)
	///////////////////////////////////////////////////////////////////////////

	BHVisualMode m_visualMode;
	BHVisualMode m_desVisualMode;


	// RealTime Velocity Mode Configuration
	double velPGain;
	double velDGain;

	long realtimeGain;
	long LCVMultiplier;

	// RealTime Torque Mode Configuration
	float torPGainFinger;
	float torDGainFinger;
	float torPGainSpread;
	float torDGainSpread;

	//double torPIDValue[4];
	//double torLastPIDValue[4];
	//double fingerLastVelError[4];

	// RealTime Position Mode Configuration
	float m_refPos[4];

	// Finger Tips
	VisualFingerTip *m_fingerTips[NUM_FINGER_TIPS];

	// Palm
	VisualPalm *m_palm;

};

#endif
