#include "VisualFingerTip.h"

#include "BHand.h"

#define PI 3.1415926535897932384626433832795
#define DEG2RAD(x) (((x) * PI) / 180.0)

const int GRIDX = 3;
const int GRIDY = 8;

const float fingerTipWidth = 0.0052f * 3; // meters
const float fingerTipLength = 0.045f;

const float wAmount = fingerTipWidth / GRIDX;

float lAmountIndex[GRIDY] = { 0.0060f, 0.0060f, 0.0060f, 0.0060f, 0.0060f, 0.0035f, 0.0035f, 0.0035f };

float pressureAngle[GRIDY] = { -5, 0, 5, 5, 20, 40, 60, 80}; // degrees (visual for finger tip begins at tip and builds right to left)

VisualFingerTip *VisualFingerTip::m_fingerTip[3] = { NULL, NULL, NULL };

VisualFingerTip::VisualFingerTip(int motor) :
	m_motor(motor), pressureMap(GRIDY, GRIDX), vert(GRIDY+1, GRIDX+1)
{
	m_fingerTip[motor] = this;

	// Initialize pressure map
	for (int r = 0; r < GRIDY; r++)
		for (int c = 0; c < GRIDX; c++)
			pressureMap(r, c) = ((float)(r + 1) / (float)GRIDY + (float)(c + 1) / (float)GRIDX) / 2.0f;

	float x = 0.010f;
	float z = 0.0f;

	// Initialize fixed vertices (that will be transformed later)
	for (int r = GRIDY; r >= 0; r--)
	{
		if (r < GRIDY)
		{
			x -= (float)cos(DEG2RAD(pressureAngle[r])) * lAmountIndex[r];
			z += (float)sin(DEG2RAD(pressureAngle[r])) * lAmountIndex[r];
		}

		for (int c = 0; c < GRIDX + 1; c++)
		{
			float y = -(((float)c) * wAmount - (wAmount * (float)GRIDX * 0.5f));

			vert(r, c).x = x;
			vert(r, c).y = y;
			vert(r, c).z = z;

			// Curve finger tip manually
			if (c == 0 || c == GRIDX)
				vert(r, c).z = vert(r, c).z - 0.0005f;

			if (r == GRIDY - 0 && (c == 0 || c == GRIDX))
				vert(r, c).x = vert(r, c).x - 0.0009f;
			if (r == GRIDY - 1 && (c == 0 || c == GRIDX))
				vert(r, c).x = vert(r, c).x - 0.0006f;
			if (r == GRIDY - 2 && (c == 0 || c == GRIDX))
				vert(r, c).x = vert(r, c).x - 0.0002f;
		}
	}
}

VisualFingerTip::~VisualFingerTip()
{
	while (!queue.empty())
		delete (VisualDataPPS *)queue.dequeueNotification();
}

VisualFingerTip* VisualFingerTip::getInstance(int motor)
{
	return m_fingerTip[motor];
}

void VisualFingerTip::init()
{
	// Will be called after OpenGL is Initialized (placeholder for now)
}

void VisualFingerTip::update(BHand *bhand)
{
	VisualDataPPS *ppsData = new VisualDataPPS;
	bhand->RTGetPPS('1' + m_motor, ppsData->data, MAX_PPS_ELEMENTS);
	queue.enqueueNotification(ppsData);
}

void VisualFingerTip::render(BMatrix<double> &trans)
{
	// Get updated pressure map
	while (!queue.empty())
	{
		VisualDataPPS *ppsData = (VisualDataPPS *)queue.dequeueNotification();
		if (queue.empty())
		{
			// Only use the most recent pressure profile sensor data
			for (int r = 0; r < GRIDY; r++)
				for (int c = 0; c < GRIDX; c++)
					pressureMap(r, c) = (float)ppsData->data[r * GRIDX + c] / 4095;
		}
		delete ppsData;
	}


	// Use immediate mode
	glEnable(GL_COLOR_MATERIAL);
	glShadeModel(GL_FLAT);
	
	// Convert to single precision floating point transformation matrix
	BMatrix<float> m(4, 4);
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			m(i, j) = (float)trans(i, j);

	BMatrix<float> temp(4, 1);
	BMatrix<float> m1(4, 1);
	m1(3, 0) = 1;

	// Transform each vertex
	BVector3<float> vertices[(GRIDY + 1) * (GRIDX + 1)];
	int i = 0;
	for (int r = 0; r < GRIDY + 1; r++)
	{
		for (int c = 0; c < GRIDX + 1; c++)
		{
			m1(0, 0) = vert(r, c).x;
			m1(1, 0) = vert(r, c).y;
			m1(2, 0) = vert(r, c).z;
			temp = m.prod(m1);
			vertices[i++] = BVector3<float>(temp(0,0), temp(1,0), temp(2,0));
		}
	}

	// Get pointers to vertices
	BVector3<float>
		*v1 = vertices,
		*v2 = vertices + 1,
		*v3 = vertices + GRIDX + 1,
		*v4 = vertices + GRIDX + 2;

	// Render fingertip
	for (int r = 0; r < GRIDY; r++)
	{
		for (int c = 0; c < GRIDX; c++)
		{
			float color = pressureMap(r, c);
			glColor3f(color, color, 0.5f + color * 0.5f);
			glBegin(GL_TRIANGLE_STRIP);
				glVertex3f(v1->x, v1->y, v1->z);
				glVertex3f(v2->x, v2->y, v2->z);
				glVertex3f(v3->x, v3->y, v3->z);
				glVertex3f(v4->x, v4->y, v4->z);
			glEnd();
			v1++; v2++; v3++; v4++;
		}
		v1++; v2++; v3++; v4++;
	}

	// Render sides of finger tip
	glColor3f(0.2f, 0.2f, 0.8f);
	BVector3<float> vo;
	v1 = vertices; // reset pointer to vertex

	m1(0, 0) = vert(0, 0).x;
	m1(1, 0) = vert(0, 0).y;
	m1(2, 0) = 0;
	temp = m.prod(m1); vo = BVector3<float>(temp(0,0), temp(1,0), temp(2,0));

	glBegin(GL_TRIANGLE_FAN);
		glVertex3f(vo.x, vo.y, vo.z);
		for (int r = 0; r < GRIDY + 1; r++)
		{
			glVertex3f(v1->x, v1->y, v1->z);
			v1 += GRIDX + 1;
		}
	glEnd();

	v1 = vertices + (GRIDX + 1) * GRIDY + GRIDX; // reset pointer to vertex

	m1(0, 0) = vert(0, GRIDX).x;
	m1(1, 0) = vert(0, GRIDX).y;
	m1(2, 0) = 0;
	temp = m.prod(m1); vo = BVector3<float>(temp(0,0), temp(1,0), temp(2,0));

	glBegin(GL_TRIANGLE_FAN);
		glVertex3f(vo.x, vo.y, vo.z);
		for (int r = GRIDY; r >= 0; r--)
		{
			glVertex3f(v1->x, v1->y, v1->z);
			v1 -= GRIDX + 1;
		}
	glEnd();

	glShadeModel(GL_SMOOTH);
	glDisable(GL_COLOR_MATERIAL);
}
