#include "VisualPalm.h"

#include "VisualData.h"

#include "BHand.h"

const int GRIDX = 7;
const int GRIDY = 4;

const float wAmount = 0.011f; // meters
const float lAmount = 0.011f;

VisualPalm *VisualPalm::m_palm = NULL;

VisualPalm::VisualPalm() :
	pressureMap(GRIDY, GRIDX), vert(GRIDY+1, GRIDX+1)
{
	m_palm = this;

	// Initialize pressure map
	for (int r = 0; r < GRIDY; r++)
		for (int c = 0; c < GRIDX; c++)
			pressureMap(r, c) = ((float)(r + 1) / (float)GRIDY + (float)(c + 1) / (float)GRIDX) / 2.0f;

	// Initialize fixed vertices (that will be transformed later)
	for (int r = 0; r < GRIDY + 1; r++)
	{
		float z = 0.0005f;

		for (int c = 0; c < GRIDX + 1; c++)
		{
			vert(r, c).x = ((float)r) * lAmount - (lAmount * (float)GRIDY * 0.5f);
			vert(r, c).y = ((float)c) * wAmount - (wAmount * (float)GRIDX * 0.5f);
			vert(r, c).z = z;
		}
	}
}

VisualPalm::~VisualPalm()
{
	while (!queue.empty())
		delete (VisualDataPPS *)queue.dequeueNotification();
}

VisualPalm* VisualPalm::getInstance()
{
	return m_palm;
}

void VisualPalm::update(BHand *bhand)
{
	VisualDataPPS *ppsData = new VisualDataPPS;
	bhand->RTGetPPS('4', ppsData->data, MAX_PPS_ELEMENTS);
	queue.enqueueNotification(ppsData);
}

void VisualPalm::render(BMatrix<double> &trans)
{
	// Get updated pressure map
	while (!queue.empty())
	{
		VisualDataPPS *ppsData = (VisualDataPPS *)queue.dequeueNotification();
		if (queue.empty())
		{
			// Only use the most recent pressure profile sensor data
			int i = 0;
			for (int r = 0; r < GRIDY; r++)
				for (int c = 0; c < GRIDX; c++)
				{
					if (r == 0 && (c == 0 || c == GRIDX - 1))
						continue;
					if (r == GRIDY - 1 && (c == 0 || c == GRIDX - 1))
						continue;
					//pressureMap(r, c) = (float)ppsData->data[i++] / 4095;
					pressureMap(r, GRIDX-1 - c) = (float)ppsData->data[i++] / 4095; // flip pressure mapping for now
				}
		}
		delete ppsData;
	}

	// Use immediate mode
	glEnable(GL_COLOR_MATERIAL);
	glShadeModel(GL_FLAT);
	
	// Convert to single precision floating point transformation matrix
	BMatrix<float> m(4, 4);
	for (int i = 0; i < 4; i++)
		for (int j = 0; j < 4; j++)
			m(i, j) = (float)trans(i, j);

	// Render the palm sensor referenced from finger 3 frame 1
	BMatrix<float> temp(4, 1);
	BMatrix<float> m1(4, 1);
	m1(3, 0) = 1;

	// Transform each vertex
	BVector3<float> vertices[(GRIDY + 1) * (GRIDX + 1)];
	int i = 0;
	for (int r = 0; r < GRIDY + 1; r++)
	{
		for (int c = 0; c < GRIDX + 1; c++)
		{
			m1(0, 0) = vert(r, c).x;
			m1(1, 0) = vert(r, c).y;
			m1(2, 0) = vert(r, c).z;
			temp = m.prod(m1);
			vertices[i++] = BVector3<float>(temp(0,0), temp(1,0), temp(2,0));
		}
	}

	// Get pointers to vertices
	BVector3<float>
		*v1 = vertices,
		*v2 = vertices + 1,
		*v3 = vertices + GRIDX + 1,
		*v4 = vertices + GRIDX + 2;

	// Render palm
	for (int r = 0; r < GRIDY; r++)
	{
		for (int c = 0; c < GRIDX; c++)
		{
			if ((r == 0 || r == GRIDY - 1) && (c == 0 || c == GRIDX - 1))
			{
				// skip cells in the corner
				v1++; v2++; v3++; v4++;
				continue;
			}

			float color = pressureMap(r, c);
			glColor3f(color, color, 0.5f + color * 0.5f);

			glBegin(GL_TRIANGLE_STRIP);
				glVertex3f(v1->x, v1->y, v1->z);
				glVertex3f(v3->x, v3->y, v3->z);
				glVertex3f(v2->x, v2->y, v2->z);
				glVertex3f(v4->x, v4->y, v4->z);
			glEnd();
			v1++; v2++; v3++; v4++;
		}
		v1++; v2++; v3++; v4++;
	}

	glShadeModel(GL_SMOOTH);
	glDisable(GL_COLOR_MATERIAL);
}
