#include "Visualizer.h"

#include <wx/wx.h>

#include "BHand.h"
#include "trackball.h"
#include "Models/VRMLModel.h"
#include "Common/Vector4.h"
#include "HandManager.h"
#include "VisualPanel.h"


#define PI 3.1415926535897932384626433832795

#ifdef LINUX
    static const float mouse_scale = 1.0;
#else
    static const float mouse_scale = 1.5;
#endif

bool hasInit = false;
bool displayTrackball = true;

float m_scale = 1.0;
int attrib_list[] = { WX_GL_RGBA, WX_GL_DOUBLEBUFFER, WX_GL_DEPTH_SIZE, 16, 0 };

float deltaRot[] = {0.0, 0.0, 0.0, 0.0};
float totalRot[] = {0.0, 0.0, 0.0, 0.0};
float rotMat[4][4];

int screenWidth;
int screenHeight;


VisualPanel *vp;



BEGIN_EVENT_TABLE(Visualizer, wxGLCanvas)
    EVT_PAINT       (Visualizer::OnPaint)
    EVT_IDLE        (Visualizer::OnIdle)
    EVT_MOUSE_EVENTS(Visualizer::OnMouse)
    EVT_ERASE_BACKGROUND(Visualizer::OnEraseBackground)
    EVT_ENTER_WINDOW(Visualizer::OnEnterWindow)
END_EVENT_TABLE()

Visualizer::Visualizer(wxWindow *parent, wxSlider* s1, wxSlider* s2, wxSlider* s3, wxSlider* s4)
	: wxGLCanvas(parent, wxID_ANY,  wxPoint(5,15), wxSize(635,635), wxDOUBLE_BORDER, wxT("GLCanvas"), attrib_list)
{
	hand_model = NULL;

	vp = (VisualPanel *)parent;
	int argc = 1;
	char* argv[] = { (char *)"\x00" };

	init = false;


	//!initialize
	glutInit(&argc, argv);

	current_x = 0.0;
	current_y = 0.0;
	old_x = -1;
	old_y = -1;
	zoom = 0.0;

	m_sliders[0] = s1;
	m_sliders[1] = s2;
	m_sliders[2] = s3;
	m_sliders[3] = s4;

	ToggleColorPickingOn(false);

	//!for display (screen resolution) size
	wxDisplaySize(&screenWidth, &screenHeight);
}

Visualizer::~Visualizer()
{
	delete hand_model;
}

//! Initialize openGL settings
void Visualizer::Init() const
{
	GLfloat light_ambient[4];
	GLfloat light_diffuse[4];
	GLfloat light_specular[4];
	GLfloat light_emissive[4];
	GLfloat light_position[4];
	GLfloat light_position2[4];

	light_ambient[0] = 0.1f;
	light_ambient[1] = 0.1f;
	light_ambient[2] = 0.1f;
	light_ambient[3] = 0.05f;

	light_diffuse[0] = 0.24f;
	light_diffuse[1] = 0.24f;
	light_diffuse[2] = 0.24f;
	light_diffuse[3] = 0.1f;

	light_specular[0] = 0.4f;
	light_specular[1] = 0.4f;
	light_specular[2] = 0.4f;
	light_specular[3] = 0.15f;

	light_emissive[0] = 0.4f;
	light_emissive[1] = 0.4f;
	light_emissive[2] = 0.4f;
	light_emissive[3] = 0.1f;

	light_position[0] = 100.0f;
	light_position[1] = 100.0f;
	light_position[2] = 100.0f;
	light_position[3] = 1.0f;

	light_position2[0] = -100.0f;
	light_position2[1] = -100.0f;
	light_position2[2] = 100.0f;
	light_position2[3] = 1.0f;

	glEnable(GL_LIGHTING);

	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glClearDepth(1.0);


	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	//!Toggle off to improve performance

	/* Enable a single OpenGL light. */
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT0, GL_EMISSION, light_emissive);
	glLightfv(GL_LIGHT0, GL_POSITION, light_position);
	glLightfv(GL_LIGHT1, GL_AMBIENT, light_ambient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light_diffuse);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light_specular);
	glLightfv(GL_LIGHT1, GL_EMISSION, light_emissive);
	glLightfv(GL_LIGHT1, GL_POSITION, light_position2);

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);

	GLfloat lmodel_twoside[1] = {GL_TRUE};
	glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);

	glClearColor(0.3f, 0.3f, 0.3f, 1.0);		/* This Will Clear The Background Color To Black */
	glShadeModel(GL_SMOOTH);


	//!@ Toggle off hints to improve performance
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);

	/* Setup the view of the cube. */
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective( /* field of view in degree */ 45.0,
	/* aspect ratio */ 1.0, // width / height
	/* Z near */ 0.1, /* Z far */ 50.0);
	//glTranslatef(0.0,-0.5,0.0);
	//glRotatef(-5,1.0,0.0,0.0);
}

void Visualizer::ShowKinematics(bool show) { hand_model->showKinematics(show); }

void Visualizer::DisplayStrain(bool display) { hand_model->displayStrain(display); }
void Visualizer::DisplayPPS(bool display) { hand_model->displayPPS(display); }


void Visualizer::LoadParts()
{
	//! make a hand model
	hand_model = new HandManager(0.0, 0.0, 0.0, 60.0, 0.0, 0.0, 1.0);
}

void Visualizer::OnPaint(wxPaintEvent& WXUNUSED(event))
{
	wxPaintDC dc(this);
	#ifndef __WXMOTIF__
	if (!GetContext()) return;
	#endif
	SetCurrent();
	// init OpenGL once, but after SetCurrent
	if (!init)
	{
		Init();
		LoadParts();
		init = true;
	}

	//clear the display
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	//set the correct camera
	UpdateCamera();
	//render the scene
	Render();
	//double-buffered so switch to the other buffer and prepare to render the next frame
	glFlush();
	SwapBuffers();

	//event.Skip();
}

//! This method is responsible to clear the background. However
//! it does nothing but consume the event. This is to avoid flashing
//! effects.
void Visualizer::OnEraseBackground(wxEraseEvent & event)
{
	//event.Skip();
}

//! Focus handling. This event handler consumes the event.
void Visualizer::OnEnterWindow(wxMouseEvent & event)
{
	event.Skip();
}

//! Scene is rendered when Idle (this is essentially the "repaint" method)
void Visualizer::OnIdle(wxIdleEvent &event)
{
	Refresh(TRUE);
	Update();
	//event.Skip();
}

void Visualizer::OnMouse(wxMouseEvent & event)
{
	if (!init)
		return;

	//! update position
	current_x = event.GetX();
	current_y = event.GetY();

	//wxPuts(wxString::Format(wxT("Visualizer::OnMouse(%d, %d)"), current_x, current_y));

	//! Trackball rotation using quaternions

	if ((event.Dragging() && event.LeftIsDown()) || event.LeftDown())
	{
		//! update position and store previous position
		int x, y;

		//! transfer from wxGLCanvas-space to absolute screen-space
		//! update position
		GetScreenPosition(&x, &y);
		current_x_absolute = (2 * ((x + current_x) / (float) screenWidth) - 1.0) * mouse_scale;
		current_y_absolute = ((2 * ((y + current_y) / (float) screenHeight) - 1.0) * -1.0) * mouse_scale;

		if (event.Dragging() && event.LeftIsDown())
		{
			if (hasInit)
			{
				//figure out the rotation in quaternions
				trackball(deltaRot, old_x_absolute, old_y_absolute, current_x_absolute, current_y_absolute);
				add_quats(deltaRot, totalRot, totalRot);
			}
			else
			{
				//just one quaternion no need to combine quaternions yet
				trackball(totalRot, old_x_absolute, old_y_absolute, current_x_absolute, current_y_absolute);
			}

			hasInit = true;
		}

		//! store previous position
		old_x_absolute = current_x_absolute;
		old_y_absolute = current_y_absolute;
	}

	//! Following color picking algorithm borrowed from:
	//link - http://gpwiki.org/index.php/OpenGL_Selection_Using_Unique_Color_IDs

	if ((event.RightDown() && event.ControlDown()) || event.RightDown())
	{
		//! Enable color picking and render the scene and store in a buffer (don't display on screen)
		ToggleColorPickingOn(true);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		Render();

		// get color information from frame buffer
		unsigned char pixel[3];

		GLint viewport[4];
		glGetIntegerv(GL_VIEWPORT, viewport);

		//! sample the color of the pixel clicked by mouse and find which object it belongs to
		glReadPixels(event.GetX(), viewport[3] - event.GetY(), 1, 1, GL_RGB, GL_UNSIGNED_BYTE, pixel);
		selected_part = hand_model->FindSelectedPart(pixel);

		//! Disable color picking
		ToggleColorPickingOn(false);

		//! Add current selection to selected objects if Ctrl is held down
		if (event.RightDown() && event.ControlDown())
		{

			bool alreadySelected = hand_model->isSelected(selected_part);
			//if model hasn't been selected yet, add it to selected objects
			if (!alreadySelected && selected_part)
			{
				selected_part->setSelected(true);
				hand_model->SelectUpdate();
				//otherwise deselect it and remove it from selected objects
			}
			else if (selected_part)
			{
				selected_part->setSelected(false);
				hand_model->DeselectUpdate();
			}

			SetFocus();
		}

	}

	//! move the selected_part accordingly

	if (event.RightIsDown())
	{
		if (event.Dragging())
			hand_model->Move(selected_part, (int)(mouse_scale * (current_x - old_x)), (int)(mouse_scale * (current_y - old_y)));
		else if (event.GetWheelRotation() > 0 && visual_interact)
			hand_model->Move(selected_part, -event.GetWheelDelta() / 7, -event.GetWheelDelta() / 7);
		else if (event.GetWheelRotation() < 0 && visual_interact)
			hand_model->Move(selected_part, event.GetWheelDelta() / 7, event.GetWheelDelta() / 7);
	}

	//! move selected parts in unison (active if interact mode is toggled)

	#ifdef LINUX

	if (event.GetWheelRotation() > 0 && visual_interact)
		hand_model->MoveSelected(0, -event.GetWheelDelta() / 7);
	else if (event.GetWheelRotation() < 0 && visual_interact)
		hand_model->MoveSelected(0, event.GetWheelDelta() / 7);

	#else

	if (event.GetWheelRotation() > 0 && visual_interact)
		hand_model->MoveSelected(0, -event.GetWheelDelta() / 5);
	else if (event.GetWheelRotation() < 0 && visual_interact)
		hand_model->MoveSelected(0, event.GetWheelDelta() / 5);

	#endif

	m_sliders[0]->SetValue((int)hand_model->getFingerRot1());
	m_sliders[1]->SetValue((int)hand_model->getFingerRot2());
	m_sliders[2]->SetValue((int)hand_model->getFingerRot3());
	m_sliders[3]->SetValue((int)hand_model->getSpreadRot());

	if (event.RightDown())
		vp->OnVisualizerRightMouseButton(false);

	if (event.RightUp())
	{
		vp->OnVisualizerRightMouseButton(true);
		selected_part = NULL;
	}

	//! camera zoom code (active if camera mode is toggled)
	if (event.GetWheelRotation() > 0 && visual_camera)
		zoom += (event.GetWheelDelta() / 2400.0);
	else if (event.GetWheelRotation() < 0 && visual_camera)
		zoom -= (event.GetWheelDelta() / 2400.0);

	//! store previous position
	old_x = current_x;
	old_y = current_y;

	event.Skip();
}

void Visualizer::setVisualizerMode(bool camera)
{
	visual_camera = camera;
	visual_interact = !camera;

	SetFocus();
}

void Visualizer::UpdateCamera()
{
	glMatrixMode(GL_MODELVIEW);

	//! Modelview matrix: reset it and translate the scene away from the viewer
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -0.5 + zoom);

	//Note: we are using trackball rotation with respect to the scene, not from the viewer's perspective.
	//      Hence, we apply the trackball rotation to the modelview matrix and not the projection matrix.
	//      You can try it out if you want to see the effect.
	build_rotmatrix(rotMat, totalRot);

	float mat[16];

	for(unsigned int i = 0; i < 16; i++)
		mat[i] = rotMat[i/4][i%4];

	glMultMatrixf(mat);
}

void Visualizer::ToggleColorPickingOn(bool flag)
{
	colorPicking = flag;
}

void Visualizer::ShowTrackball(bool flag)
{
	displayTrackball = flag;
}

void Visualizer::RefreshOpenGL(float f1, float f2, float f3, float spread)
{
	hand_model->RefreshLocation(f1,f2,f3,spread);

	m_sliders[0]->SetValue((int)f1);
	m_sliders[1]->SetValue((int)f2);
	m_sliders[2]->SetValue((int)f3);
	m_sliders[3]->SetValue((int)spread);
}


float Visualizer::getFinger1Rot() { return hand_model->getFingerRot1(); }
float Visualizer::getFinger2Rot() { return hand_model->getFingerRot2(); }
float Visualizer::getFinger3Rot() { return hand_model->getFingerRot3(); }
float Visualizer::getSpreadRot()  { return hand_model->getSpreadRot();  }


//! renders the scene and trackball
void Visualizer::Render() const
{
	if (colorPicking)
	{
		glDisable(GL_LIGHTING);
		glDisable(GL_DITHER);
		glDisable (GL_BLEND);
		glDisable (GL_TEXTURE_1D);
		glDisable (GL_TEXTURE_2D);
		glShadeModel (GL_FLAT);
	}

	// clear the scene and make sure we are making changes to the modelview matrix
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glColor3f(0.2f, 0.4f, 1.0f);

	// draw trackball
	if (displayTrackball)
		glutWireSphere(0.2, 20, 20);

	// draw the hand
	//glPushAttrib(GL_ENABLE_BIT);
	hand_model->DrawHand(colorPicking);
	//glPopAttrib();


	if (colorPicking)
	{
		glEnable(GL_LIGHTING);
		glEnable(GL_DITHER);
		glEnable (GL_BLEND);
		glEnable (GL_TEXTURE_1D);
		glEnable (GL_TEXTURE_2D);
		glShadeModel (GL_SMOOTH);
	}
}
