///////////////////////////////////////////////////////////
//                                                       //
//                                                       //
//    Monitor Strain Gage Program                        //
//    v1.0 990803 bz - Created Program                   //
//                                                       //
///////////////////////////////////////////////////////////


#include <stdio.h>
#ifndef LINUX
#include <conio.h>
#endif
#include "BHand.h"

#ifdef LINUX
    #include <sys/time.h> // Needed for ms time.

    struct HiResTimer {
        void Start(void) { gettimeofday(&mTimeStart, NULL); };
        double GetElapsedInMilliSecs(void)
        {
            struct timeval mTimeNow;
            gettimeofday(&mTimeNow, NULL);
            double duration = (double)(mTimeNow.tv_sec - mTimeStart.tv_sec) * 1000 + (double)(mTimeNow.tv_usec - mTimeStart.tv_usec) / 1000;
            return duration;
        }

        struct timeval mTimeStart;
    };

#else
    #include <Windows.h>

    struct HiResTimer {
        void Start(void) { QueryPerformanceCounter(&mTimeStart); };
        double GetElapsedInMilliSecs(void)
        {
            LARGE_INTEGER mTimeNow;
            QueryPerformanceCounter(&mTimeNow);
            LARGE_INTEGER freq;
            QueryPerformanceFrequency(&freq);
            double duration = 1000 * (double)(mTimeNow.QuadPart-mTimeStart.QuadPart)/(double)freq.QuadPart;
            return duration;
        }

        LARGE_INTEGER mTimeStart;
    };

#endif

BHand  bh;           // Handles all hand communication
int    value;        // Hand parameter obtained with Get
int    result;       // Return value (error) of all BHand calls

int    szbuffer;     // Size of data buffers
//int*   pdata[4][8];  // Pointers to data buffers
struct HiResTimer tmr;


///////////////////////////////////////////////////////////
// Error Handler - called whenever result!=0
// Feel free to customize this function

void Error()
{
	char myStr[100];

	sprintf(myStr, "err %d", result);
	bh.Command(myStr, myStr);
	printf( "\n\n\t\tERROR %s\n\n", myStr );
	//printf( "Press any key to exit..." );

	//myChar = getchar();

	exit(0);
}


///////////////////////////////////////////////////////////
//  Initialize hand, set timeouts and baud rate

void Initialize()
{
	int port;

	printf("\n\n\r\t\tInitializing Software...");
	printf("\n\n\r\t\tPlease enter desired serial port> ");
	scanf("%d", &port);
	printf("\n");

	if (result=bh.InitSoftware(port,THREAD_PRIORITY_TIME_CRITICAL))
		Error();

	if (result=bh.ComSetTimeouts(100,15000,100,5000))
		Error();

	if (result=bh.Baud(9600))
		Error();

	printf( "\r\t\t\tInitializing Hand...    " );

	if (result=bh.InitHand(""))
		Error();
}


///////////////////////////////////////////////////////////
//  Execute commands, return 1 if interrupted with a key

int Before()
{
	if (result=bh.Command("T"))
		Error();
	if (_kbhit())
		{ _getch(); return 1; }

	return 0;
}


///////////////////////////////////////////////////////////
//  Set parameters, allocate data buffers, load files

void PrepareRealTime()
{
	szbuffer = 500;

	/*for (int m = 0; m < 4; m++)
		for (int v = 0; v < 8; v++)
		{
			pdata[m][v] = new int[szbuffer];
			memset((void*)pdata[m][v], 0, szbuffer*sizeof(int));
		}*/

	if (result=bh.RTSetFlags("123", 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0))
		Error();
}


///////////////////////////////////////////////////////////
//  Run RealRime loop, return 1 if interrupted with a key

int RunRealTime()
{
	double var[4][7];
	unsigned int time;

	bh.RTStart("123");
	tmr.Start();
	bh.RTUpdate();

	printf( "\r\t\t\tPress Any Key to Abort...\n\n" );
	printf("\t\tFinger 1\tFinger 2\tFinger 3\n");

	while (!_kbhit())
	{
		time = (unsigned int)tmr.GetElapsedInMilliSecs();

		if (time > 200) //Display strain every 200 milliseconds
		{
			tmr.Start();
			for (int m = 0; m < 3; m++)
				var[m][2] = bh.RTGetStrain( m+'1' );

			printf("\r\t\t  %.0f  \t\t  %.0f  \t\t  %.0f  ", var[0][2], var[1][2], var[2][2]); fflush(stdout);
		}

		bh.RTUpdate();
	}

	bh.RTAbort();

	printf("\n\n");

	_getch();
	return 0;

}


///////////////////////////////////////////////////////////
//  Save all buffers into a text file

/*void SaveData(char* name)
{
	FILE* fp = fopen(name, "wt");
	if( !fp )
		return;

	fprintf(fp, "Finger 1\tFinger 2\tFinger 3\n");

	for( int r=0; r<szbuffer; r++ )
	{
		int v=2; //for( int v=0; v<8; v++ )
		{
			for( int m=0; m<3; m++ )
				fprintf(fp, "%d\t", pdata[m][v][r]);
			fprintf(fp, "  ");
		}
		fprintf(fp, "\n");
	}

	fclose(fp);
}*/


void PrintBanner()
{
#ifdef LINUX
	system("clear");
#else
	system("cls");
#endif
	printf("\n\t\t-------------------------------------------");
	printf("\n\t\t|                                         |");
	printf("\n\t\t|     Barrett Technology, Inc.            |");
	printf("\n\t\t|     Monitor Strain Gage Program v1.0    |");
	printf("\n\t\t|                                         |");
	printf("\n\t\t-------------------------------------------");
}



///////////////////////////////////////////////////////////
//  Main function - initialize, execute

int main()
{
	PrintBanner();

	Initialize();

	if (Before())
		return 0;

	PrepareRealTime();

	if (RunRealTime())
		return 0;

	return 0;
}
